class FT1D:
    def __init__(self, *data):
        if not all([isinstance(i, (int, float, complex)) for i in data]):
            raise ValueError("The input data must all be numerical values")
        n = len(data)
        if n & (n - 1):
            n = 2 ** n.bit_length()
            data = list(data) + [0] * (n - len(data))
        self.data = data

    def __repr__(self):
        return self.__class__.__name__ + str(self.data)

    def fft(self):
        from cmath import exp

        def inner(data):
            n = len(data)
            if n <= 1:
                return data
            even = inner(data[0::2])
            odd = inner(data[1::2])
            t = [exp(-6.283185307179586j * k / n) * odd[k] for k in range(n // 2)]
            return [even[k] + t[k] for k in range(n // 2)] + [even[k] - t[k] for k in range(n // 2)]

        self.data = inner(self.data)

    def ifft(self):
        from cmath import exp

        def inner(data):
            n = len(data)
            if n <= 1:
                return data
            even = inner(data[0::2])
            odd = inner(data[1::2])
            t = [exp(6.283185307179586j * k / n) * odd[k] for k in range(n // 2)]
            return [even[k] + t[k] for k in range(n // 2)] + [even[k] - t[k] for k in range(n // 2)]

        self.data = [k / len(self.data) for k in inner(self.data)]



from .arrays import Array as __Array


def apply(a, func, rtype=None):
    def inner(data):
        return [inner(item) for item in data] if isinstance(data, list) else func(data)

    if isinstance(a, __Array):
        if callable(func):
            if rtype is None:
                rtype = type(a)
            return rtype(inner(a.data))
        else:
            raise TypeError("The 'func' argument must be callable, got {} instead".format(type(func)))
    else:
        raise TypeError("The 'a' argument must be an instance of Array, got {} instead".format(type(a)))


def base_ufunc(*arrays, func, args=(), rtype=None):
    _type = str(type(func))
    if not all([isinstance(item, __Array) for item in arrays]) or not (
            _type.startswith("<function ") or _type.startswith("<class ")):
        raise TypeError("The input parameter type is incorrect")
    if len(set([item.shape for item in arrays])) != 1:
        raise ValueError("All arrays must have the same shape")
    data = [item.data for item in arrays]

    def inner(*arrs):
        if isinstance(arrs[0], list):
            _copy = []
            for item in zip(*arrs):
                _copy.append(inner(*item))
            return _copy
        else:
            return func(*arrs, *args) if args else func(*arrs)

    if rtype is None:
        rtype = type(arrays[0])
    return rtype(inner(*data))


def ufunc_helper(x, y, func):
    if isinstance(x, __Array):
        if isinstance(y, __Array):
            return base_ufunc(x, y, func=func)
        else:
            return base_ufunc(x, func=func, args=[y])
    else:
        if isinstance(y, __Array):
            return base_ufunc(y, func=lambda b, a: func(a, b), args=[x])
        else:
            return func(x, y)


def add(x, y):
    def inner(_x, _y):
        return _x + _y

    return ufunc_helper(x, y, inner)


def subtract(x, y):
    def inner(_x, _y):
        return _x - _y

    return ufunc_helper(x, y, inner)


def multiply(x, y):
    def inner(_x, _y):
        return _x * _y

    return ufunc_helper(x, y, inner)


def divide(x, y):
    def inner(_x, _y):
        return _x / _y

    return ufunc_helper(x, y, inner)


def floor_divide(x, y):
    def inner(_x, _y):
        return _x // _y

    return ufunc_helper(x, y, inner)


def modulo(x, y):
    def inner(_x, _y):
        return _x % _y

    return ufunc_helper(x, y, inner)


def power(x, y, m=None):
    from .arrays import full

    def inner(_x, _y, _m=None):
        try:
            return pow(_x, _y, _m)
        except ValueError:
            return pow(_x, _y)

    is_array = [isinstance(x, __Array), isinstance(y, __Array), isinstance(m, __Array)]
    if not any(is_array):
        return pow(x, y, m)
    first = (x, y, m)[is_array.index(True)]
    shape = first.shape
    if not is_array[0]:
        x = full(shape, x)
    if not is_array[1]:
        y = full(shape, y)
    if not is_array[2]:
        m = full(shape, m)
    return base_ufunc(x, y, m, func=inner, rtype=type(first))


def gt(x, y):
    return ufunc_helper(x, y, lambda a, b: a > b)


def ge(x, y):
    return ufunc_helper(x, y, lambda a, b: a >= b)


def eq(x, y):
    return ufunc_helper(x, y, lambda a, b: a == b)


def le(x, y):
    return ufunc_helper(x, y, lambda a, b: a <= b)


def lt(x, y):
    return ufunc_helper(x, y, lambda a, b: a < b)


def ne(x, y):
    return ufunc_helper(x, y, lambda a, b: a != b)



def pprint_matrix(matrix, style="numpy", output=True):
    if str(type(matrix)) != "<class 'pypynum.Matrix.Matrix'>":
        raise TypeError("The input must be of type 'pypynum.Matrix.Matrix'")
    style = style.strip().lower()
    supported_styles = ["numpy", "mpmath", "sympy", "borderless", "numbered"]
    if style not in supported_styles:
        raise ValueError("Unsupported style '{}'. Supported styles are: {}".format(style, supported_styles))
    max_length = max([len(str(item)) for row in matrix for item in row])
    if style == "numbered":
        max_length = max(max_length, len(str(matrix.cols)))
    format_str = "{:" + str(max_length) + "}"
    row_strings = []
    separator = "  " if style in ["mpmath", "sympy"] else " "
    for row in matrix:
        formatted_row = separator.join([format_str.format(item) for item in row])
        row_strings.append(formatted_row)
    result = ""
    if style == "numpy":
        result = str(matrix)
    elif style == "mpmath":
        result = "\n".join(["[" + row_str + "]" for row_str in row_strings])
    elif style == "sympy":
        border_length = max_length + 2
        column_count = matrix.cols
        middle_border = "⎢" + " " * (border_length * column_count - 2) + "⎥"
        end = len(row_strings) - 1
        if end < 1:
            result = "[" + row_strings[0] + "]"
        else:
            result = "⎡" + row_strings[0] + "⎤"
            for i, row_str in enumerate(row_strings[1:end]):
                result += "".join(["\n", middle_border, "\n", "⎢", row_str, "⎥"])
            result += "".join(["\n", middle_border, "\n", "⎣", row_strings[-1], "⎦"])
    elif style == "borderless":
        result = "\n".join(row_strings)
    elif style == "numbered":
        row_number_width, column_number_width = matrix.rows, matrix.cols
        row_number_format = "{:>" + str(len(str(row_number_width))) + "}"
        column_number_format = "{:^" + str(max_length) + "}"
        column_header = " " * len(str(row_number_width)) + " | " + " ".join(
            [column_number_format.format(str(j + 1)) for j in range(column_number_width)])
        for i, row in enumerate(row_strings):
            row_string = row_number_format.format((i + 1)) + " | " + row
            row_strings[i] = row_string
        result = "".join([column_header, "\n", "-" * len(row_strings[0]), "\n", "\n".join(row_strings)])
    if output:
        print(result)
    else:
        return result



from .arrays import Array
from .types import ShapeError

MatchError = ShapeError("The shapes of two vectors do not match")


class Vector(Array):
    """
    It is a vector class that supports basic operations,
    as well as functions such as norm and all angles.
    :param data: An array in the form of a list
    :param check: Check the rationality of the input array
    """

    def __init__(self, data, check=True):
        super().__init__(data, check)
        if len(self.shape) != 1:
            raise ShapeError("Vectors can only be one-dimensional in shape")
        self.len = len(data)

    def __add__(self, other):
        if self.len != other.len:
            raise MatchError
        return Vector([self.data[i] + other.data[i] for i in range(self.len)])

    def __sub__(self, other):
        if self.len != other.len:
            raise MatchError
        return Vector([self.data[i] - other.data[i] for i in range(self.len)])

    def __mul__(self, other):
        if isinstance(other, Vector):
            if self.len != other.len:
                raise MatchError
            return Vector([self.data[i] * other.data[i] for i in range(self.len)])
        else:
            return Vector([i * other for i in self.data])

    def __matmul__(self, other):
        if self.len != other.len:
            raise MatchError
        return sum([self.data[i] * other.data[i] for i in range(self.len)])

    def __abs__(self):
        return self.norm()

    def __pos__(self):
        return Vector([item for item in self.data])

    def __neg__(self):
        return Vector([-item for item in self.data])

    def norm(self, p=2):
        if p == 0:
            return len([item for item in self.data if item != 0])
        elif p == 1:
            return sum([abs(item) for item in self.data])
        elif p == 2:
            return sum([i ** 2 for i in self.data]) ** 0.5
        elif p == float("inf"):
            return max([abs(item) for item in self.data])
        else:
            return sum([i ** p for i in self.data]) ** (1 / p)

    def normalize(self):
        norm = self.norm()
        return Vector([i / norm for i in self.data])

    def angles(self, axes=None):
        from math import acos
        if axes is None or axes == [] or axes == ():
            result = []
            for a in range(self.len):
                axis = Vector([1 if p == a else 0 for p in range(self.len)])
                result.append(acos(self @ axis / (self.norm() * axis.norm())))
        elif int(abs(axes)) == axes:
            axis = Vector([1 if p == axes else 0 for p in range(self.len)])
            result = acos(self @ axis / (self.norm() * axis.norm()))
        elif isinstance(axes, (list, tuple)):
            result = []
            for a in axes:
                axis = Vector([1 if p == a else 0 for p in range(self.len)])
                result.append(acos(self @ axis / (self.norm() * axis.norm())))
        else:
            raise TypeError("The axes parameter can only be None, natural number, list, or tuple")
        return result


def vec(data):
    return Vector(data)


del Array



ContentError = ValueError("The content of the string is invalid")
ROMAN_VALUES = [1000, 900, 500, 400, 100, 90, 50, 40, 10, 9, 5, 4, 1]
ROMAN_SYMBOLS = ["M", "CM", "D", "CD", "C", "XC", "L", "XL", "X", "IX", "V", "IV", "I"]


def int2words(integer: int) -> str:
    """
    introduction
    ==========
    Convert integers into natural language.

    example
    ==========
    >>> int2words(4294967296)
    'four billion two hundred and ninety-four million nine hundred and sixty-seven thousand two hundred and ninety-six'
    >>>
    :param integer: integer
    :return:
    """
    if not isinstance(integer, int):
        raise TypeError("The input must be an integer")
    if integer == 0:
        return "zero"
    if integer < 0:
        return "negative " + int2words(abs(integer))
    ones = ["", "one", "two", "three", "four", "five", "six", "seven", "eight", "nine", "ten",
            "eleven", "twelve", "thirteen", "fourteen", "fifteen", "sixteen", "seventeen", "eighteen", "nineteen"]
    tens = ["", "", "twenty", "thirty", "forty", "fifty", "sixty", "seventy", "eighty", "ninety"]

    def two_digit(num):
        if num < 20:
            return ones[num] if num > 0 else ""
        ten, one = divmod(num, 10)
        return tens[ten] + ("-" if one else "") + ones[one]

    def three_digit(num):
        if num == 0:
            return ""
        hundred, rem = divmod(num, 100)
        return ones[hundred] + " hundred" + (" and " if rem else "") + two_digit(rem)

    big_units = ["quadragintillion", "novemtrigintillion", "octrigintillion", "septrigintillion", "sextrigintillion",
                 "quintrigintillion", "quattuortrigintillion", "trestrigintillion", "duotrigintillion",
                 "untrigintillion", "trigintillion", "novemvigintillion", "octovigintillion", "septenvigintillion",
                 "sexvigintillion", "quinvigintillion", "quattuorvigintillion", "trevigintillion", "duovigintillion",
                 "unvigintillion", "vigintillion", "novemdecillion", "octodecillion", "septendecillion", "sexdecillion",
                 "quindecillion", "quattuordecillion", "tredecillion", "duodecillion", "undecillion", "decillion",
                 "nonillion", "octillion", "septillion", "sextillion", "quintillion", "quadrillion", "trillion",
                 "billion", "million", "thousand"]
    unit_value = 10 ** (3 * len(big_units))
    result = []
    for name in big_units:
        if integer >= unit_value:
            value, integer = divmod(integer, unit_value)
            result.append(int2words(value) + " " + name)
        unit_value //= 1000
    if integer > 0:
        words = two_digit(integer) if integer < 100 else three_digit(integer)
        result.append(words)
    return " ".join(result)


def str2int(string: str) -> int:
    """
    introduction
    ==========
    Converts strings to integers and supports parsing ordinary floating-point numbers
    and scientific notation floating-point numbers.

    example
    ==========
    >>> str2int("0.123456789e+5")
    12345
    >>>
    :param string: string
    :return:
    """
    for item in string:
        if item not in "0123456789e.+-":
            raise ContentError
    length = 0
    string = string.lower()
    if "e" in string:
        if string.count("e") != 1:
            raise ContentError
        parts = string.split("e")
        integer_part = parts[0]
        exponent_part = parts[1]
        if not exponent_part:
            raise ContentError
        if "." in integer_part:
            if integer_part.count(".") != 1:
                raise ContentError
            integer_part, length = integer_part.replace(".", ""), len(integer_part.split(".")[1])
        integer_value = int(integer_part)
        exponent_value = int(exponent_part)
        if exponent_value > 0:
            integer_value *= 10 ** exponent_value
        elif exponent_value < 0:
            div = divmod(integer_value, 10 ** (-exponent_value))
            integer_value = div[0] if not div[1] or div[0] >= 0 else div[0] + 1
        if integer_value and length:
            div = divmod(integer_value, 10 ** length)
            return div[0] if not div[1] or div[0] >= 0 else div[0] + 1
        else:
            return integer_value
    else:
        if "." in string:
            if string.count(".") != 1:
                raise ContentError
            integer_part = string.split(".")[0]
            return int(integer_part) if integer_part else 0
        else:
            return int(string)


def int2roman(integer: int, overline: bool = True) -> str:
    """
    introduction
    ==========
    Convert natural numbers to Roman numerals and support with or without an overline.

    example
    ==========
    >>> int2roman(12345)
    'X̄ĪĪCCCXLV'
    >>>
    :param integer: integer
    :param overline: bool
    :return:
    """
    if not isinstance(integer, int):
        if isinstance(integer, float):
            integer = round(integer)
        else:
            raise TypeError("The number to be converted can only be a natural number")
    if integer < 0:
        raise ValueError("The number to be converted can only be a natural number")

    def int2roman_helper(number):
        roman = ""
        i = 0
        while number > 0:
            tmp = number // ROMAN_VALUES[i]
            roman += ROMAN_SYMBOLS[i] * tmp
            number -= ROMAN_VALUES[i] * tmp
            i += 1
        return roman

    def roman_overline(num):
        return "̄".join(int2roman_helper(num)) + "̄"

    if overline and integer >= 4000:
        thousands = integer // 1000
        remainder = integer % 1000
    else:
        thousands = 0
        remainder = integer
    roman_num = ""
    if thousands > 0:
        roman_num += roman_overline(thousands)
    roman_num += int2roman_helper(remainder)
    return roman_num


def roman2int(roman_num: str) -> int:
    """
    introduction
    ==========
    Convert Roman numerals to natural numbers and support the presence or absence of an overline.

    example
    ==========
    >>> roman2int("X̄ĪĪCCCXLV")
    12345
    >>>
    :param roman_num: string
    :return:
    """

    def roman2int_helper(number):
        part = 0
        last = 10000
        i = 0
        while i < len(number):
            if number[i:i + 2] in ROMAN_SYMBOLS:
                value = ROMAN_VALUES[ROMAN_SYMBOLS.index(number[i:i + 2])]
                if value > last:
                    raise ContentError
                else:
                    last = value
                part += value
                i += 2
            elif number[i] in ROMAN_SYMBOLS:
                value = ROMAN_VALUES[ROMAN_SYMBOLS.index(number[i])]
                if value > last:
                    raise ContentError
                else:
                    last = value
                part += value
                i += 1
            else:
                raise ContentError
        return part

    if "̄" in roman_num:
        index = roman_num.rfind("̄")
        high = roman_num[:index + 1]
        over = high[1::2]
        if over.count("̄") != len(over):
            raise ContentError
        return roman2int_helper(high.replace("̄", "")) * 1000 + roman2int_helper(roman_num[index + 1:])
    else:
        return roman2int_helper(roman_num)


def float2fraction(number: float, mixed: bool = False, error: float = 1e-15) -> tuple:
    """
    introduction
    ==========
    Convert floating-point numbers to fractions and support mixed numbers and false fractions.

    example
    ==========
    >>> float2fraction(3.141592653589793, False, 1e-6)
    (355, 113)
    >>>
    :param number: float
    :param mixed: bool
    :param error: float
    :return:
    """
    if number < 0:
        number = -number
        flag = True
    else:
        flag = False
    whole = int(number)
    number -= whole
    numerator = 0
    denominator = 1
    while abs(number - float(numerator) / float(denominator)) > error:
        if number > float(numerator) / float(denominator):
            numerator += 1
        else:
            denominator += 1
    if flag:
        whole = -whole
        numerator = -numerator
    return (whole, numerator, denominator) if mixed else (whole * denominator + numerator, denominator)


def split_float(s: str) -> tuple:
    """
    Introduction
    ==========
    Decompose a floating-point number string into its sign, integer part, fractional part, and exponent part.

    Example
    ==========
    >>> split_float("-123.456e+7")
    ('-', '123', '456', '+7')
    >>>
    :param s: The string representation of the floating-point number to be decomposed.
    :return: A tuple containing the sign, integer part, fractional part, and exponent part of the input string.
    :raises ValueError: If the input string cannot be converted to a float, represents a NaN (Not a Number),
        or represents an infinity.
    """
    sign, integer, fraction, exponent = "", "", "", ""
    s = str(s).strip()
    try:
        floating = float(s)
    except ValueError:
        raise ValueError("The input string cannot be converted to a float")
    if floating != floating:
        raise ValueError("The input string represents a NaN (Not a Number)")
    if s in ("inf", "-inf"):
        raise ValueError("The input string represents an infinity")
    if s[0] in ["+", "-"]:
        sign, s = s[0], s[1:]
    if "e" in s or "E" in s:
        exp_index = s.index("e") if "e" in s else s.index("E")
        exponent, s = s[exp_index + 1:], s[:exp_index]
    if "." in s:
        dot_index = s.index(".")
        integer, fraction = s[:dot_index], s[dot_index + 1:].rstrip("0")
    else:
        integer = s
    return sign, integer, fraction, exponent


def parse_float(s: str) -> tuple:
    """
    Introduction
    ==========
    Parse a floating-point number string and return the sign bit, exponent, and mantissa.

    Example
    ==========
    >>> parse_float("-123.456e+7")
    ('1', '4', '123456')
    >>>
    :param s: The string representation of the floating-point number to be parsed.
    :return: A tuple containing the sign bit (0 for positive, 1 for negative), the exponent, and the mantissa.
    :raises ValueError: If the input string is not a valid floating-point number string.
    """
    sign, integer, fraction, exponent = split_float(s)
    sign = "1" if sign == "-" else "0"
    new_exponent = int(exponent) if exponent else 0
    if fraction:
        new_exponent -= len(fraction)
    mantissa = integer + fraction or "0"
    new_mantissa = mantissa.lstrip("0")
    if not new_mantissa:
        new_mantissa = "0"
    new_exponent -= len(mantissa) - len(new_mantissa)
    return sign, str(new_exponent), new_mantissa


if __name__ == "__main__":
    import doctest

    doctest.testmod(verbose=True)



from .types import config


class Polynomial:
    def __init__(self, terms=None):
        if terms is None:
            self.terms = []
        else:
            try:
                self.terms = sorted([tuple(item) if len(item) == 2 else None for item in terms if item[1] != 0])
            except TypeError:
                raise TypeError("The input polynomial must be a sequence consisting of degree-coefficient pairs")

    def add_term(self, degree, coefficient):
        if coefficient == 0:
            return

        def binary_search(terms, deg):
            low, high = 0, len(terms) - 1
            while low <= high:
                mid = (low + high) // 2
                if terms[mid][0] == deg:
                    return mid
                elif terms[mid][0] < deg:
                    low = mid + 1
                else:
                    high = mid - 1
            return low

        insert_index = binary_search(self.terms, degree)
        if insert_index < len(self.terms) and self.terms[insert_index][0] == degree:
            try:
                self.terms[insert_index] = (degree, self.terms[insert_index][1] + coefficient)
            except OverflowError:
                self.terms[insert_index] = (degree, int(self.terms[insert_index][1]) + int(coefficient))
        else:
            self.terms.insert(insert_index, (degree, coefficient))

    def is_zero(self):
        return not self.terms

    def gcd(self, other):
        temp = self
        while not other.is_zero():
            temp, other = other, temp % other
        return temp

    def lcm(self, other):
        gcd_poly = self.gcd(other)
        if gcd_poly.is_zero():
            raise ValueError("Cannot compute LCM of polynomials with zero GCD")
        return self * other // gcd_poly

    def roots(self, real_roots=True, tolerance=1e-15):
        from .equations import poly_eq
        all_roots = poly_eq(self.coeffs(True))
        if real_roots:
            return [(1 if root.real > 0 else -1) * abs(root) for root in all_roots if abs(root.imag) < tolerance]
        return all_roots

    def evaluate(self, x):
        return sum([coefficient * x ** degree for degree, coefficient in self.terms])

    def degs(self, reverse=False):
        terms = reversed(self.terms) if reverse else self.terms
        return [deg for deg, _ in terms]

    def coeffs(self, reverse=False):
        max_deg = self.degree()
        coeffs = [0] * (max_deg + 1)
        for deg, coeff in self.terms:
            coeffs[deg] = coeff
        if reverse:
            coeffs.reverse()
        return coeffs

    def sqrt(self):
        deg, coeff = self.terms[-1]
        new = deg // 2, coeff ** 0.5
        root = Polynomial([new])
        temp = Polynomial([new])
        rem = self - root * root
        for _ in range(deg):
            temp.add_term(*temp.terms[0])
            try:
                root.add_term(*(rem // temp).terms[-1])
            except IndexError:
                break
            temp.add_term(*root.terms[0])
            rem = rem - temp * Polynomial([temp.terms[0]])
        return root, rem

    def remove0terms(self):
        self.terms = [(deg, coeff) for deg, coeff in self.terms if coeff != 0]

    def degree(self):
        if self.terms:
            return self.terms[-1][0]
        return -1

    def lead_coeff(self):
        if self.terms:
            return self.terms[-1][1]
        return 0

    def deriv(self):
        derivative_terms = [(deg - 1, deg * coeff) for deg, coeff in self.terms if deg > 0]
        return Polynomial(derivative_terms)

    def integ(self, constant=0):
        integrated_terms = [(deg + 1, coeff / (deg + 1)) for deg, coeff in self.terms]
        if constant != 0:
            integrated_terms.append((0, constant))
        return Polynomial(integrated_terms)

    def latex(self):
        return self.__repr__(True)

    def __repr__(self, use_latex=False):
        from .chars import int2superscript
        use_latex = config.use_latex or use_latex
        use_unicode = config.use_unicode
        if not self.terms:
            return "0"
        result = ""
        for i, (degree, coefficient) in enumerate(self.terms):
            if i != 0:
                if coefficient < 0:
                    result += " - "
                else:
                    result += " + "
            elif coefficient < 0:
                result += "-"
            if degree == 0:
                result += str(abs(coefficient))
            elif abs(coefficient) != 1:
                result += str(abs(coefficient))
            if degree > 0:
                if degree > 1:
                    if use_latex:
                        result += "x^{" + str(degree) + "}"
                    elif use_unicode:
                        result += "x" + int2superscript(str(degree))
                    else:
                        result += "x^" + str(degree)
                else:
                    result += "x"
        return result

    def __pos__(self):
        return Polynomial([(degree, coefficient) for degree, coefficient in self.terms])

    def __neg__(self):
        return Polynomial([(degree, -coefficient) for degree, coefficient in self.terms])

    def __add__(self, other):
        if isinstance(other, (int, float)):
            return Polynomial([(deg, coeff + other if deg == 0 else coeff) for deg, coeff in self.terms])
        result = Polynomial()
        i, j = 0, 0
        while i < len(self.terms) or j < len(other.terms):
            if i == len(self.terms):
                result.add_term(other.terms[j][0], other.terms[j][1])
                j += 1
            elif j == len(other.terms):
                result.add_term(self.terms[i][0], self.terms[i][1])
                i += 1
            else:
                degree_self, coefficient_self = self.terms[i]
                degree_other, coefficient_other = other.terms[j]
                if degree_self < degree_other:
                    result.add_term(degree_self, coefficient_self)
                    i += 1
                elif degree_self > degree_other:
                    result.add_term(degree_other, coefficient_other)
                    j += 1
                else:
                    try:
                        result.add_term(degree_self, coefficient_self + coefficient_other)
                    except OverflowError:
                        result.add_term(degree_self, int(coefficient_self) + int(coefficient_other))
                    i += 1
                    j += 1
        return result

    def __sub__(self, other):
        if isinstance(other, (int, float)):
            return Polynomial([(deg, coeff - other if deg == 0 else coeff) for deg, coeff in self.terms])
        negative_other = -other
        return self + negative_other

    def __mul__(self, other):
        if isinstance(other, (int, float)):
            return Polynomial([(deg, coeff * other) for deg, coeff in self.terms])
        result_terms = {}
        for degree_self, coeff_self in self.terms:
            for degree_other, coeff_other in other.terms:
                degree_product = degree_self + degree_other
                coeff_product = coeff_self * coeff_other
                if degree_product in result_terms:
                    result_terms[degree_product] = result_terms[degree_product] + coeff_product
                else:
                    result_terms[degree_product] = coeff_product
        return Polynomial(sorted(result_terms.items()))

    def __truediv__(self, other):
        if isinstance(other, (int, float)):
            if other == 0:
                raise ValueError("Cannot divide by zero")
            return Polynomial([(deg, coeff / other) for deg, coeff in self.terms])
        if not other.terms or other.lead_coeff() == 0:
            raise ValueError("Cannot divide by zero polynomial")
        quotient = Polynomial()
        remainder = Polynomial(self.terms)
        other_degree = other.degree()
        other_lead_coeff = other.lead_coeff()
        while remainder.degree() >= other_degree:
            quotient_degree = remainder.degree() - other_degree
            try:
                quotient_coefficient = (remainder.lead_coeff() / other_lead_coeff)
            except OverflowError:
                quotient_coefficient = (remainder.lead_coeff() // other_lead_coeff)
            quotient_term = Polynomial([(quotient_degree, quotient_coefficient)])
            quotient.add_term(quotient_degree, quotient_coefficient)
            remainder = remainder - quotient_term * other
        return quotient, remainder

    def __floordiv__(self, other):
        if isinstance(other, (int, float)):
            return Polynomial([(deg, coeff // other) for deg, coeff in self.terms])
        return (self / other)[0]

    def __mod__(self, other):
        if isinstance(other, (int, float)):
            return Polynomial([(deg, coeff % other) for deg, coeff in self.terms])
        return (self / other)[1]

    def __divmod__(self, other):
        if isinstance(other, (int, float)):
            if other == 0:
                raise ValueError("Cannot divide or modulo by zero")
            int_part_terms = []
            mod_part_terms = []
            for deg, coeff in self.terms:
                int_coeff, mod_coeff = divmod(coeff, other)
                if int_coeff != 0 or deg == 0:
                    int_part_terms.append((deg, int_coeff))
                if mod_coeff != 0:
                    mod_part_terms.append((deg, mod_coeff))
            int_part = Polynomial(int_part_terms)
            mod_part = Polynomial(mod_part_terms)
            return int_part, mod_part
        else:
            return self / other

    def __int__(self):
        self.terms = [(degree, int(coefficient)) for degree, coefficient in self.terms]
        return 0

    def __float__(self):
        self.terms = [(degree, float(coefficient)) for degree, coefficient in self.terms]
        return 0.0

    def __round__(self, n=None):
        return Polynomial([(degree, round(coefficient, n)) for degree, coefficient in self.terms])

    def __pow__(self, power, modulo=None):
        if power == 0:
            return Polynomial([(0, 1)])
        result = Polynomial([(0, 1)])
        for _ in range(power):
            result = result * self
        if modulo:
            result = result % modulo
        return result

    def stationaries(self, tolerance=1e-15):
        derivative = self.deriv()
        return derivative.roots(tolerance=tolerance)

    def saddles(self, tolerance=1e-15):
        first_derivative = self.deriv()
        second_derivative = first_derivative.deriv()
        stationary_points = self.stationaries(tolerance=tolerance)
        saddle_points = []
        for point in stationary_points:
            if abs(second_derivative.evaluate(point)) > tolerance and first_derivative.evaluate(
                    point) * second_derivative.evaluate(point) < 0:
                saddle_points.append(point)
        return saddle_points

    def minima(self, tolerance=1e-15):
        second_derivative = self.deriv().deriv()
        stationary_points = self.stationaries(tolerance=tolerance)
        minimums = []
        for point in stationary_points:
            if second_derivative.evaluate(point) > 0:
                minimums.append(point)
        return minimums

    def maxima(self, tolerance=1e-15):
        second_derivative = self.deriv().deriv()
        stationary_points = self.stationaries(tolerance=tolerance)
        maximums = []
        for point in stationary_points:
            if second_derivative.evaluate(point) < 0:
                maximums.append(point)
        return maximums

    def inflections(self, tolerance=1e-15):
        second_derivative = self.deriv().deriv()
        third_derivative = second_derivative.deriv()
        potential_inflections = second_derivative.roots(tolerance=tolerance)
        inflection_points = [x for x in potential_inflections if abs(third_derivative.evaluate(x)) > tolerance]
        return inflection_points


def poly(terms=None):
    return Polynomial(terms)


def from_coeffs(coeffs):
    return Polynomial([(degree, coefficient) for degree, coefficient in enumerate(coeffs) if coefficient != 0])


def from_coords(coords):
    from .regs import poly_reg
    return from_coeffs(reversed(poly_reg(*zip(*coords))))


def legpoly(n, single=True):
    x = Polynomial([(1, 1)])
    p = [Polynomial([(0, 1)]), x]
    for i in range(1, n):
        m = Polynomial([(0, (2 * i + 1) / (i + 1))]) * x * p[i] - Polynomial([(0, i / (i + 1))]) * p[i - 1]
        p.append(m)
    return p[n] if single else p


def leggauss(n):
    polynomial = legpoly(n)
    roots = polynomial.roots(False)
    weights = []
    for node in roots:
        factor1 = 1 - node ** 2
        derivative_poly = polynomial.deriv()
        derivative_value = derivative_poly.evaluate(node)
        weight = 2 / (factor1 * derivative_value ** 2)
        weights.append(weight)
    return roots, weights


def chebpoly(n, single=True):
    result = [Polynomial([(0, 1)]), Polynomial([(1, 1)])]
    multiple = Polynomial([(1, 2)])
    for _ in range(n - 1):
        result.append(multiple * result[-1] - result[-2])
    return result[n] if single else result


def chebgauss(n):
    from math import cos
    pi_over_n = 3.141592653589793 / n
    pi_over_2n = pi_over_n / 2
    roots = [cos((2 * i - 1) * pi_over_2n) for i in range(1, n + 1)]
    weights = [pi_over_n for _ in range(n)]
    return roots, weights


def lagpoly(n, single=True):
    result = [Polynomial([(0, 1)]), Polynomial([(0, 1), (1, -1)])]
    for i in range(2, n + 1):
        result.append((Polynomial([(0, 2 * i - 1), (1, -1)]) * result[-1] - Polynomial([(0, i - 1)]) * result[-2]) / i)
    return result[n] if single else result


def laggauss(n):
    polynomial, next_polynomial = lagpoly(n + 1, False)[-2:]
    roots = polynomial.roots()
    weights = []
    const = (n + 1) ** 2
    for node in roots:
        next_polynomial_value = next_polynomial.evaluate(node)
        weight = node / const * next_polynomial_value ** 2
        weights.append(weight)
    return roots, weights



from .types import Any, Callable, Union, real


class Series:
    def __init__(self, data: Any = None, index: Any = None) -> None:
        if data is None:
            self.data = []
            self.index = []
        else:
            self.data = data
            self.index = index

    def __repr__(self) -> str:
        def string(obj):
            return obj.__str__().__repr__()[1:-1]

        escaped_indices = tuple(map(string, self.index))
        escaped_data = tuple(map(string, self.data))
        max_index_length = max(map(len, escaped_indices))
        max_data_length = max(map(len, escaped_data))
        format_str = "{{:<{}}}    {{:>{}}}".format(max_index_length, max_data_length)
        return "\n".join([format_str.format(index, data) for index, data in zip(escaped_indices, escaped_data)])

    def describe(self, percentiles: Any = (0.25, 0.5, 0.75), interpolation: str = "linear", dof: int = 1) -> "Series":
        from .maths import quantile
        data = self.__dropna(None)
        if all([isinstance(x, (int, float)) for x in self.data]):
            count = len(data)
            mean = sum(data) / count
            std = (sum([(x - mean) ** 2 for x in data]) / (count - dof)) ** 0.5
            min_val = min(data)
            max_val = max(data)
            percentiles = sorted(set(percentiles))

            def func(q):
                return quantile(data, q, interpolation, True)

            data = [count, mean, std, min_val] + list(map(func, percentiles)) + [max_val]
            index = ["count", "mean", "std", "min"] + [str(p * 100) + "%" for p in percentiles] + ["max"]
            return Series(data, index)
        else:
            count = len(data)
            unique = len(set(data))
            top = max(data, key=data.count)
            freq = data.count(top)
            data = [count, unique, top, freq]
            index = ["count", "unique", "top", "freq"]
            return Series(data, index)

    def count(self, skipna: bool = True) -> int:
        return len(self.__dropna(None) if skipna else self.data)

    def mean(self, skipna: bool = True) -> real:
        from .maths import mean
        data = self.__dropna(None) if skipna else self.data
        return mean(data)

    def var(self, dof: int = 1, skipna: bool = True) -> real:
        from .maths import var
        data = self.__dropna(None) if skipna else self.data
        return var(data, dof)

    def std(self, dof: int = 1, skipna: bool = True) -> real:
        from .maths import std
        data = self.__dropna(None) if skipna else self.data
        return std(data, dof)

    def min(self, skipna: bool = True) -> real:
        data = self.__dropna(None) if skipna else self.data
        return min(data)

    def max(self, skipna: bool = True) -> real:
        data = self.__dropna(None) if skipna else self.data
        return max(data)

    def quantile(self, q: real, interpolation: str = "linear", skipna: bool = True) -> real:
        from .maths import quantile
        data = self.__dropna(None) if skipna else self.data
        return quantile(data, q, interpolation)

    def __dropna(self, ignore_index: Union[bool, None] = False) -> Union["Series", tuple]:
        filtered_pairs = [(x, idx) for x, idx in zip(self.data, self.index) if self.__notna(x)]
        new_data, new_index = zip(*filtered_pairs) if filtered_pairs else ((), ())
        if ignore_index is None:
            return new_data
        if ignore_index:
            new_index = range(len(new_data))
        return Series(new_data, new_index)

    def dropna(self, ignore_index: bool = False) -> "Series":
        return self.__dropna(bool(ignore_index))

    def __getattribute__(self, item: Any) -> Any:
        if item == "data" or item == "index":
            return list(super().__getattribute__(item))
        else:
            if item in self.index:
                data = [x for x, idx in zip(self.data, self.index) if idx == item]
                length = len(data)
                if length == 1:
                    return data[0]
                else:
                    return Series(data, [item for _ in range(length)])
            else:
                return super().__getattribute__(item)

    def __setattr__(self, key: Any, value: Any) -> None:
        if key == "data":
            if isinstance(value, (list, tuple)):
                if hasattr(self, "index") and len(value) != len(self.index):
                    raise ValueError("Length of 'data' does not match length of 'index'.")
                super().__setattr__(key, list(value))
            else:
                raise TypeError("'data' must be a list or tuple.")
        elif key == "index":
            if value is None:
                super().__setattr__(key, list(range(len(self.data))))
            elif isinstance(value, (list, tuple)):
                if len(self.data) != len(value):
                    raise ValueError("Length of 'index' does not match length of 'data'.")
                super().__setattr__(key, list(value))
            else:
                raise TypeError("'index' must be a list or tuple.")
        else:
            if key in self.index:
                data = [x if idx != key else value for x, idx in zip(self.data, self.index)]
                super().__setattr__("data", data)
            else:
                super().__setattr__(key, value)

    def apply(self, func: Callable) -> "Series":
        return Series(map(func, self.data), self.index)

    def head(self, n: int = 5) -> "Series":
        return Series(self.data[:n], self.index[:n])

    def tail(self, n: int = 5) -> "Series":
        return Series(self.data[-n:], self.index[-n:])

    def abs(self) -> "Series":
        return Series(map(abs, self.data), self.index)

    def sum(self, skipna: bool = True) -> real:
        from math import fsum
        data = self.__dropna(None) if skipna else self.data
        return fsum(data)

    def idxmax(self, skipna: bool = True) -> Any:
        data = self.__dropna(None) if skipna else self.data
        return self.index[data.index(max(data))]

    def idxmin(self, skipna: bool = True) -> Any:
        data = self.__dropna(None) if skipna else self.data
        return self.index[data.index(min(data))]

    def isna(self) -> "Series":
        return Series(map(self.__isna, self.data), self.index)

    def notna(self) -> "Series":
        return Series(map(self.__notna, self.data), self.index)

    def fillna(self, value: Any) -> "Series":
        filled_data = [value if self.__isna(x) else x for x in self.data]
        return Series(filled_data, self.index)

    def replace(self, to_replace: Any, value: Any) -> "Series":
        return Series([value if x == to_replace else x for x in self.data], self.index)

    def copy(self) -> "Series":
        return Series(self.data, self.index)

    def __sort(self, items, ascending=True, na_position="last"):
        def sort_key(item):
            value = item[1]
            isna = self.__isna(value)
            return isna if bool(ascending) is (na_position == "last") else not isna, item[0] if isna else value

        if na_position not in ["first", "last"]:
            raise ValueError("'na_position' must be either 'first' or 'last'")
        sorted_items = sorted(enumerate(items), key=sort_key, reverse=not ascending)
        return tuple(zip(*sorted_items))

    def sort_index(self, ascending: bool = True, na_position: str = "last", ignore_index: bool = False) -> "Series":
        sorted_positions, sorted_indices = self.__sort(self.index, ascending, na_position)
        sorted_values = map(self.data.__getitem__, sorted_positions)
        new_index = range(len(self.index)) if ignore_index else sorted_indices
        return Series(sorted_values, new_index)

    def sort_values(self, ascending: bool = True, na_position: str = "last", ignore_index: bool = False) -> "Series":
        sorted_positions, sorted_values = self.__sort(self.data, ascending, na_position)
        new_index = range(len(self.index)) if ignore_index else map(self.index.__getitem__, sorted_positions)
        return Series(sorted_values, new_index)

    @staticmethod
    def __isna(value: Any) -> bool:
        return value is None or isinstance(value, float) and value != value

    @staticmethod
    def __notna(value: Any) -> bool:
        return not Series.__isna(value)

    def cov(self, other: "Series", min_periods: int = None, dof: int = 1) -> real:
        from .maths import cov
        data1 = self.__dropna(None)
        data2 = other.__dropna(None)
        if len(data1) < min_periods or len(data2) < min_periods:
            return float("nan")
        return cov(data1, data2, dof)

    def corr(self, other: "Series", min_periods: int = None) -> real:
        from .maths import corr_coeff
        data1 = self.__dropna(None)
        data2 = other.__dropna(None)
        if len(data1) < min_periods or len(data2) < min_periods:
            return float("nan")
        return corr_coeff(data1, data2)



def lin_eq(left: list, right: list) -> list:
    from .matrices import mat
    try:
        return (mat(left).inv() @ mat([right]).t()).t()[0]
    except ValueError:
        return [float("inf")] * len(right)


def poly_eq(coefficients: list) -> list:
    from .matrices import eigen, mat
    p = [_ / coefficients[0] for _ in coefficients[1:]]
    return sorted(eigen(mat([[-p[i] if j == 0 else 1 if i + 1 == j else 0 for j in range(len(p))]
                             for i in range(len(p))]))[0].diag().t()[0], key=lambda c: (c.real, c.imag))



from .kernels import matmul2x2kernel, matmul3x3kernel, matmul4x4kernel
from .tools import fast_pow
from .types import Union, arr, real


def pascal_triangle(n: int) -> list:
    triangle = [[1]]
    for i in range(1, n + 1):
        new_row = [1]
        for j in range(1, i):
            new_row.append(triangle[i - 1][j - 1] + triangle[i - 1][j])
        new_row.append(1)
        triangle.append(new_row)
    return triangle


def bell_triangle(n: int) -> list:
    triangle = [[1]]
    for i in range(1, n + 1):
        new_row = [triangle[i - 1][-1]]
        for j in range(1, i + 1):
            new_row.append(new_row[j - 1] + triangle[i - 1][j - 1])
        triangle.append(new_row)
    return triangle


def stirling1(n: int) -> list:
    triangle = [[1]]
    for i in range(1, n + 1):
        new_row = [0]
        for j in range(1, i):
            new_row.append((i - 1) * triangle[i - 1][j] + triangle[i - 1][j - 1])
        new_row.append(triangle[i - 1][i - 1])
        triangle.append(new_row)
    return triangle


def stirling2(n: int) -> list:
    triangle = [[1]]
    for i in range(1, n + 1):
        new_row = [0]
        for j in range(1, i):
            new_row.append(j * triangle[i - 1][j] + triangle[i - 1][j - 1])
        new_row.append(triangle[i - 1][i - 1])
        triangle.append(new_row)
    return triangle


def farey(n: int) -> list:
    a2, b2 = None, None
    if n == 0:
        return []
    elif n == 1:
        return [(0, 1), (1, 1)]
    else:
        f = farey(n - 1)
        result = []
        for i in range(len(f) - 1):
            a1, b1 = f[i]
            a2, b2 = f[i + 1]
            result.append((a1, b1))
            temp = (a1 + a2, b1 + b2)
            if temp[0] <= temp[1] <= n:
                result.append(temp)
        result.append((a2, b2))
        return result


def fibonacci(n: int, single: bool = True) -> Union[int, list]:
    if isinstance(n, float):
        n = int(n)
    if not single:
        if n == 0:
            return [0]
        res = [0, 1]
        for i in range(2, n + 1):
            res.append(res[i - 1] + res[i - 2])
        return res
    if n <= 1:
        return max(n, 0)
    a = [[1, 1], [1, 0]]
    res = fast_pow(a, n, [[1, 0], [0, 1]], matmul2x2kernel)
    return res[0][1]


def pell(n: int, single: bool = True) -> Union[int, list]:
    if isinstance(n, float):
        n = int(n)
    if not single:
        if n == 0:
            return [0]
        res = [0, 1]
        for i in range(2, n + 1):
            res.append(2 * res[i - 1] + res[i - 2])
        return res
    if n <= 1:
        return max(n, 0)
    a = [[2, 1], [1, 0]]
    res = fast_pow(a, n, [[1, 0], [0, 1]], matmul2x2kernel)
    return res[0][1]


def lucas(n: int, single: bool = True) -> Union[int, list]:
    if isinstance(n, float):
        n = int(n)
    if not single:
        if n == 0:
            return [2]
        res = [2, 1]
        for i in range(2, n + 1):
            res.append(res[i - 1] + res[i - 2])
        return res
    if n == 0:
        return 2
    if n == 1:
        return 1
    a = [[1, 1], [1, 0]]
    res = fast_pow(a, n, [[1, 0], [0, 1]], matmul2x2kernel)
    return res[0][0] + res[1][1]


def tribonacci(n: int, single: bool = True) -> Union[int, list]:
    if isinstance(n, float):
        n = int(n)
    res = [0, 1, 1]
    if not single:
        if n < 3:
            return res[:n]
        for i in range(3, n + 1):
            res.append(res[i - 1] + res[i - 2] + res[i - 3])
        return res[:n] if n < 3 else res
    if n < 3:
        return res[n]
    a = [[1, 1, 1], [1, 0, 0], [0, 1, 0]]
    res = fast_pow(a, n - 1, [[1, 0, 0], [0, 1, 0], [0, 0, 1]], matmul3x3kernel)
    return res[0][0]


def tetranacci(n: int, single: bool = True) -> Union[int, list]:
    if isinstance(n, float):
        n = int(n)
    res = [0, 1, 1, 2]
    if not single:
        if n < 4:
            return res[:n]
        for i in range(4, n + 1):
            res.append(res[i - 1] + res[i - 2] + res[i - 3] + res[i - 4])
        return res
    if n < 4:
        return res[n]
    a = [[1, 1, 1, 1], [1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 1, 0]]
    res = fast_pow(a, n - 1, [[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 1, 0], [0, 0, 0, 1]], matmul4x4kernel)
    return res[0][0]


def catalan(n: int, single: bool = True) -> Union[int, list]:
    try:
        from math import comb
    except ImportError:
        from .maths import combination as comb
    if isinstance(n, float):
        n = int(n)
    if not single:
        if n == 0:
            return [1]
        res = [1]
        for i in range(1, n + 1):
            res.append(res[i - 1] * (4 * i - 2) // (i + 1))
        return res
    return comb(2 * n, n) // (n + 1)


def bernoulli(n: int, single: bool = True) -> arr:
    from fractions import Fraction
    try:
        from math import comb
    except ImportError:
        from .maths import combination as comb
    if isinstance(n, float):
        n = int(n)
    result = [1, Fraction(-0.5)]
    if not single and n == 0:
        return [(1, 1)]
    for m in range(2, n + 1):
        if m & 1:
            b = 0
        else:
            s = sum([comb(m + 1, i) * result[i] for i in range(m)])
            b = -s / (m + 1)
        result.append(b)
    return result[n].as_integer_ratio() if single else [n.as_integer_ratio() for n in result]


def recaman(n: int, single: bool = True) -> Union[int, list]:
    if isinstance(n, float):
        n = int(n)
    seq = []
    for i in range(n + 1):
        if i == 0:
            x = 0
        else:
            x = seq[i - 1] - i
        if x >= 0 and x not in seq:
            seq += [x]
        else:
            seq += [seq[i - 1] + i]
    return seq[n] if single else seq


def sylvester(n: int, single: bool = True) -> Union[int, list]:
    if isinstance(n, float):
        n = int(n)
    result = [2]
    a = 2
    for _ in range(n):
        a = a * a - a + 1
        result.append(a)
    return result[n] if single else result


def padovan(n: int, single: bool = True) -> Union[int, list]:
    if isinstance(n, float):
        n = int(n)
    result = [1, 0, 0]
    if not single and n < len(result):
        return result[:n + 1]
    for i in range(3, n + 1):
        result.append(result[i - 2] + result[i - 3])
    return result[n] if single else result


def perrin(n: int, single: bool = True) -> Union[int, list]:
    if isinstance(n, float):
        n = int(n)
    result = [3, 0, 2]
    if not single and n < len(result):
        return result[:n + 1]
    for i in range(3, n + 1):
        result.append(result[i - 2] + result[i - 3])
    return result[n] if single else result


def pelllucas(n: int, single: bool = True) -> Union[int, list]:
    if isinstance(n, float):
        n = int(n)
    if not single:
        if n == 0:
            return [2]
        res = [2, 2]
        for i in range(2, n + 1):
            res.append(2 * res[i - 1] + res[i - 2])
        return res
    if n <= 1:
        return 2
    a = [[2, 1], [1, 0]]
    res = fast_pow(a, n, [[1, 0], [0, 1]], matmul2x2kernel)
    return res[0][0] + res[1][1]


def arithmetic_sequence(*, a1: real = None, an: real = None, d: real = None, n: real = None, s: real = None) -> dict:
    if [a1, an, d, n, s].count(None) != 2:
        raise ValueError("Must provide exactly three valid parameters")

    def solve_quadratic_equation(a, b, c):
        delta = b ** 2 - 4 * a * c
        if delta > 0:
            x1 = (-b + delta ** 0.5) / (2 * a)
            x2 = (-b - delta ** 0.5) / (2 * a)
            return [x1, x2]
        elif delta == 0:
            x = -b / (2 * a)
            return [x]
        else:
            return []

    if any([not isinstance(val, (int, float, type(None))) for val in [a1, an, d, n, s]]):
        raise ValueError("The input parameters must all be real numbers")
    if a1 is not None and an is not None and d is not None:
        n = (an - a1) / d + 1
        s = n / 2 * (2 * a1 + (n - 1) * d)
    elif a1 is not None and an is not None and n is not None:
        d = (an - a1) / (n - 1)
        s = n / 2 * (a1 + an)
    elif a1 is not None and an is not None and s is not None:
        n = s * 2 / (a1 + an)
        d = (an - a1) / (n - 1)
    elif a1 is not None and d is not None and n is not None:
        an = a1 + (n - 1) * d
        s = n / 2 * (2 * a1 + (n - 1) * d)
    elif a1 is not None and d is not None and s is not None:
        n = [item for item in solve_quadratic_equation(d, 2 * a1 - d, -2 * s) if item > 0]
        if len(n) == 1:
            n = n[0]
            an = a1 + (n - 1) * d
    elif a1 is not None and n is not None and s is not None:
        an = 2 * s / n - a1
        d = (an - a1) / (n - 1)
    elif an is not None and d is not None and n is not None:
        a1 = an - (n - 1) * d
        s = n / 2 * (a1 + an)
    elif an is not None and d is not None and s is not None:
        n = [item for item in solve_quadratic_equation(-d, 2 * an + d, -2 * s) if item > 0]
        if len(n) == 1:
            n = n[0]
            a1 = an - (n - 1) * d
    elif an is not None and n is not None and s is not None:
        d = (an * n - s) / (n * (n - 1) / 2)
        a1 = an - (n - 1) * d
    elif d is not None and n is not None and s is not None:
        a1 = (s / n) - (d * (n - 1)) / 2
        an = a1 + (n - 1) * d
    if isinstance(n, list):
        n = sorted(n)
    return {"a1": a1, "an": an, "d": d, "n": n, "s": s}


def geometric_sequence(*, a1: real = None, an: real = None, r: real = None, n: real = None, s: real = None) -> dict:
    from math import log
    if [a1, an, r, n, s].count(None) != 2:
        raise ValueError("Must provide exactly three valid parameters")
    if any([not isinstance(val, (int, float, type(None))) for val in [a1, an, r, n, s]]):
        raise ValueError("The input parameters must all be real numbers")
    if a1 is not None and an is not None and r is not None:
        n = log(an / a1, r) + 1
        s = a1 * (r ** n - 1) / (r - 1) if r != 1 else a1 * n
    elif a1 is not None and an is not None and n is not None:
        r = (an / a1) ** (1 / (n - 1))
        s = a1 * (r ** n - 1) / (r - 1) if r != 1 else a1 * n
    elif a1 is not None and an is not None and s is not None:
        r = (s - a1) / (s - an)
        n = log(an / a1, r) + 1
    elif a1 is not None and r is not None and n is not None:
        an = a1 * r ** (n - 1)
        s = a1 * (r ** n - 1) / (r - 1) if r != 1 else a1 * n
    elif a1 is not None and r is not None and s is not None:
        an = (s * r - s + a1) / r
        n = log(an / a1, r) + 1
    elif a1 is not None and n is not None and s is not None:
        an = NotImplemented
        r = NotImplemented
    elif an is not None and r is not None and n is not None:
        a1 = an / r ** (n - 1)
        s = a1 * (r ** n - 1) / (r - 1) if r != 1 else a1 * n
    elif an is not None and r is not None and s is not None:
        a1 = s * (1 - r) + an * r
        n = log(an / a1, r) + 1
    elif an is not None and n is not None and s is not None:
        a1 = NotImplemented
        r = NotImplemented
    elif r is not None and n is not None and s is not None:
        a1 = s * (r - 1) / (r ** n - 1)
        an = a1 * r ** (n - 1)
    if isinstance(n, list):
        n = sorted(n)
    return {"a1": a1, "an": an, "r": r, "n": n, "s": s}



OPERATORS = ["**", "*", "//", "/", "%", "+", "-"]
BASIC = "%()*+-./0123456789"
ENGLISH = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz"
GREEK = "ΑΒΓΔΕΖΗΘΙΚΛΜΝΞΟΠΡΣΤΥΦΧΨΩαβγδεζηθικλμνξοπρστυφχψω"
VALID = BASIC + ENGLISH + GREEK


def parse_expr(expr: str) -> list:
    expr = expr.replace(" ", "")
    if any([item not in VALID for item in expr]):
        raise ValueError("Characters other than Arabic numerals, English letters, Greek letters, operators, "
                         "decimal points, and parentheses cannot appear in expressions")
    depth = 0
    pointer = 0
    result = []
    for index in range(len(expr)):
        if expr[index] == "(":
            if depth == 0:
                result.append(index)
            depth += 1
        elif expr[index] == ")":
            if depth == 0:
                raise ValueError("The parentheses in the expression are not paired")
            depth -= 1
            if depth == 0:
                sub = parse_expr(expr[result[-1] + 1:index])
                if not sub:
                    raise ValueError("The inside of the parentheses in an expression cannot be empty")
                result[-1] = sub
        elif depth == 0 and expr[index] in OPERATORS:
            number = expr[pointer:index]
            if number and "(" not in number:
                if number.count(".") > 1:
                    raise ValueError("Syntax error in expression")
                nan = True
                for item in number:
                    if nan and item.isdigit():
                        nan = False
                    elif not nan and not (item == "." or item.isdigit()):
                        raise NameError("The name of the algebra is invalid")
                result.append(number)
            if index != 0:
                pointer = index + 1
            else:
                pointer = index
            if expr[index] in "*/" and expr[index] == expr[index - 1]:
                if result[-1] in OPERATORS:
                    raise ValueError("Syntax error in expression")
                result.append(expr[index] * 2)
            elif index != 0 and expr[index] != expr[index + 1]:
                if result[-1] in OPERATORS:
                    raise ValueError("Syntax error in expression")
                result.append(expr[index])
    number = expr[pointer:]
    if number and "(" not in number:
        result.append(number)
    if depth != 0:
        raise ValueError("The parentheses in the expression are not paired")
    return result if len(result) != 1 else result[0]


# TODO 表达式展开
# TODO 表达式化简
# TODO 符号微分
# TODO 符号积分
...



from .types import Any, Callable, Union, arr, ite, real


def frange(start: real, stop: real, step: float = 1.0) -> list:
    """
    Introduction
    ==========
    Float range

    Example
    ==========
    >>> frange(0, 10, 1.5)
    [0.0, 1.5, 3.0, 4.5, 6.0, 7.5, 9.0]
    >>>
    :param start: integer | float
    :param stop: integer | float
    :param step: float
    :return:
    """
    if isinstance(step, int):
        step = float(step)
    result = [_ * step for _ in range(int(start / step), int(stop / step + 1))]
    return result


def linspace(start: real, stop: real, number: int) -> list:
    """
    Introduction
    ==========
    Linear space

    Example
    ==========
    >>> linspace(2, 3, 4)
    [2.0, 2.3333333333333335, 2.6666666666666665, 3.0]
    >>>
    :param start: integer | float
    :param stop: integer | float
    :param number: integer
    :return:
    """
    if number == 1:
        return [start]
    elif number == 2:
        return [start, stop]
    else:
        step = (stop - start) / (number - 1)
        return [start + i * step for i in range(number)]


def geomspace(start: real, stop: real, number: int) -> list:
    """
    Introduction
    ==========
    Geometric space

    Example
    ==========
    >>> geomspace(2, 3, 4)
    [2.0, 2.2894284851066637, 2.620741394208897, 3.0]
    >>>
    :param start: integer | float
    :param stop: integer | float
    :param number: integer
    :return:
    """
    if number == 1:
        return [start]
    elif number == 2:
        return [start, stop]
    else:
        step = (stop / start) ** (1 / (number - 1))
        return [start * step ** i for i in range(number)]


def dedup(iterable: ite) -> ite:
    """
    Introduction
    ==========
    Data deduplication

    Example
    ==========
    >>> dedup(["T", "h", "i", "s", " ", "i", "s", " ", "a", " ", "l", "i", "s", "t"])
    ['T', 'h', 'i', 's', ' ', 'a', 'l', 't']
    >>>
    :param iterable: list | tuple | string
    :return:
    """
    result = []
    for item in iterable:
        if item not in result:
            result.append(item)
    if isinstance(iterable, list):
        return result
    elif isinstance(iterable, tuple):
        return tuple(result)
    elif isinstance(iterable, str):
        return "".join(result)
    else:
        raise TypeError("Iterable can only be a list, tuple or str")


def classify(array: arr) -> dict:
    """
    Introduction
    ==========
    Data classification

    Example
    ==========
    >>> classify((1, 2.3, 4 + 5j, 6 - 7j, 8.9, 0))
    {<class 'int'>: [1, 0], <class 'float'>: [2.3, 8.9], <class 'complex'>: [(4+5j), (6-7j)]}
    >>>
    :param array: list | tuple
    :return:
    """
    result = {}
    for item in array:
        _type = type(item)
        if _type in result:
            result[_type].append(item)
        else:
            result[_type] = [item]
    return result


def split(iterable: ite, key: arr, retain: bool = False) -> list:
    """
    Introduction
    ==========
    Data splitting

    Example
    ==========
    >>> split((1, 2, 2, 3, 3, 2, 2, 1, 2, 3, 4, 5, 6, 7, 8), [1, 3, 5], retain=True)
    [(), (1,), (2, 2), (3,), (), (3,), (2, 2), (1,), (2,), (3,), (4,), (5,), (6, 7, 8)]
    >>>
    :param iterable: list | tuple
    :param key: list | tuple
    :param retain: bool
    :return:
    """
    key = dedup(key)
    if not isinstance(retain, bool):
        raise TypeError("Only Boolean values can be used to determine whether to retain the key")
    if not isinstance(key, (list, tuple)):
        raise TypeError("The parameter 'key' can only be a list or tuple")
    result = []
    pointer = 0
    if isinstance(iterable, str):
        if key == [""]:
            raise ValueError("When iterable is a string, the key cannot be a list of empty strings")
        while True:
            indexes = {}
            for k in key:
                index = iterable.find(k, pointer)
                if index != -1 and index not in indexes:
                    indexes[index] = k
            if indexes:
                index = min(indexes)
                result.append(iterable[pointer:index])
                pointer = index
                if retain is True:
                    result.append(indexes[index])
                pointer += len(indexes[index])
            else:
                break
        result.append(iterable[pointer:])
    elif isinstance(iterable, (list, tuple)):
        for item in range(len(iterable)):
            if iterable[item] in key:
                result.append(iterable[pointer:item])
                if isinstance(iterable, (list, tuple)):
                    pointer = item + 1
                if retain is True:
                    result.append(iterable[item:item + 1])
        result.append(iterable[pointer:])
    else:
        raise TypeError("Iterable can only be a list, tuple or str")
    return result


def primality(n: int, iter_num: int = 10) -> bool:
    """
    Introduction
    ==========
    This function checks if a number is a prime using the Miller-Rabin primality test,
    which is a probabilistic test to determine if a number is a probable prime.

    Example
    ==========
    >>> primality(2 ** 4423 - 1)
    True
    >>>
    :param n: The number to be tested for primality.
    :param iter_num: The number of iterations for the test, which affects the accuracy.
        More iterations reduce the probability of falsely identifying a composite number as prime.
    :return: True if the number is a probable prime, False otherwise.
    """
    from random import randint
    if n == 2:
        return True
    elif n & 1 == 0 or n < 2:
        return False
    m, s = n - 1, 0
    while m & 1 == 0:
        m >>= 1
        s += 1
    for _ in range(iter_num):
        b = pow(randint(2, n - 1), m, n)
        if b == 1 or b == n - 1:
            continue
        for __ in range(s - 1):
            b = b * b % n
            if b == n - 1:
                break
        else:
            return False
    return True


def primes(limit: int) -> list:
    """
    Introduction
    ==========
    This function generates a list of prime numbers up to a specified limit.

    Example
    ==========
    >>> primes(100)
    [2, 3, 5, 7, 11, 13, 17, 19, 23, 29, 31, 37, 41, 43, 47, 53, 59, 61, 67, 71, 73, 79, 83, 89, 97]
    >>>
    :param limit: The upper limit to which prime numbers are to be found (inclusive).
    :return: A list of prime numbers up to the specified limit.
    """
    result = list(range(limit + 1))
    result[1] = 0
    p = 2
    s = 4
    while s <= limit:
        if result[p]:
            for i in range(s, limit + 1, p):
                result[i] = 0
        p += 1
        s = p * p
    return [x for x in result if x != 0]


def semiprimes(limit: int) -> list:
    """
    Introduction
    ==========
    This function generates a list of semiprime numbers up to a specified limit.
    A semiprime is a natural number that is the product of two prime numbers.

    Example
    ==========
    >>> semiprimes(64)
    [4, 6, 9, 10, 14, 15, 21, 22, 25, 26, 33, 34, 35, 38, 39, 46, 49, 51, 55, 57, 58, 62]
    >>>
    :param limit: The upper limit to which semiprime numbers are to be found (inclusive).
    :return: A list of semiprime numbers up to the specified limit.
    """
    ps = primes(limit // 2)
    length = len(ps)
    result = []
    for i in range(length):
        first = ps[i]
        maximum = limit // first
        for j in range(i, length):
            second = ps[j]
            if second > maximum:
                break
            result.append(first * second)
    return sorted(result)


def twinprimes(limit: int) -> list:
    """
    Introduction
    ==========
    This function generates a list of twin prime pairs up to a specified limit.
    Twin primes are a pair of prime numbers that have a difference of two.

    Example
    ==========
    >>> twinprimes(100)
    [(3, 5), (5, 7), (11, 13), (17, 19), (29, 31), (41, 43), (59, 61), (71, 73)]
    >>>
    :param limit: The upper limit to which twin prime pairs are to be found (inclusive).
    :return: A list of twin prime pairs up to the specified limit.
    """
    ps = primes(limit)
    return [(p, q) for p, q in zip(ps, ps[1:]) if q - p == 2]


def prime_factors(integer: int, dictionary: bool = False, pollard_rho: bool = True) -> Union[list, dict]:
    """
    Introduction
    ==========
    This function computes the prime factors of a given integer. It can return the factors
    as a list or a dictionary, and it can use the Pollard's rho algorithm for factorization.

    Example
    ==========
    >>> prime_factors(2305567963945518424753102147331756070)
    [2, 3, 5, 7, 11, 13, 17, 19, 23, 29, 31, 37, 41, 43, 47, 53, 59, 61, 67, 71, 73, 79, 83, 89, 97]
    >>>
    :param integer: The integer to be factorized.
    :param dictionary: A boolean indicating whether to return the factors as a dictionary with the prime factors as
        keys and their counts as values.
    :param pollard_rho: A boolean indicating whether to use Pollard's rho algorithm for factorization. If False,
        a simple trial division method is used.
    :return: A list of prime factors or a dictionary with prime factors and their counts, depending on the
        'dictionary' parameter.
    """
    from math import gcd
    from random import randint

    def normal(n):
        factors = []
        divisor = 2
        while n > 1:
            tmp = divmod(n, divisor)
            if tmp[1]:
                divisor += 1
            else:
                factors.append(divisor)
                n = tmp[0]
        return factors

    def pollard(x, c):
        i, k = 1, 2
        x0 = randint(0, x)
        y = x0
        while 1:
            i += 1
            x0 = (x0 * x0 % x + c) % x
            d = gcd(y - x0, x)
            if d != 1 and d != x:
                return d
            if y == x0:
                return x
            if i == k:
                y = x0
                k += k

    def recursion(n):
        result = []
        if primality(n):
            return [n]
        p = n
        while p >= n:
            p = pollard(p, randint(1, n - 1))
        result.extend(recursion(p))
        result.extend(recursion(n // p))
        return result

    if integer > 0:
        output = recursion(integer) if pollard_rho else normal(integer)
        if dictionary:
            return {number: output.count(number) for number in set(output)}
        else:
            return sorted(output)
    else:
        return {} if dictionary else []


def magic_square(n: int) -> list:
    """
    Introduction
    ==========
    Generate a magic square of a specified order

    Example
    ==========
    >>> magic_square(4)
    [[16, 2, 3, 13], [5, 11, 10, 8], [9, 7, 6, 12], [4, 14, 15, 1]]
    >>>
    :param n: integer
    :return:
    """
    if not isinstance(n, int) or n < 3:
        raise ValueError("The order of the magic square must be an integer greater than or equal to three")
    result = [[0] * n for _ in range(n)]

    def odd():
        i, j = 0, n // 2
        for m in range(1, n * n + 1):
            result[i][j] = m
            new_i, new_j = (i - 1) % n, (j + 1) % n
            if result[new_i][new_j] != 0:
                new_i, new_j = (i + 1) % n, j
            i, j = new_i, new_j

    def single_even():
        half = n // 2
        left = magic_square(half)
        x = n ** 2 // 4
        k = (n - 2) // 4
        for i in range(half):
            for j in range(half):
                result[i][j] = left[i][j]
                result[i][j + half] = left[i][j] + 2 * x
                result[i + half][j] = left[i][j] + 3 * x
                result[i + half][j + half] = left[i][j] + x
        for i in range(half):
            for j in range(n):
                if j > (half + k + 1):
                    result[i][j], result[i + half][j] = result[i + half][j], result[i][j]
                if j < k:
                    result[i][j], result[i + half][j] = result[i + half][j], result[i][j]
        result[k][0], result[k + half][0] = result[k + half][0], result[k][0]
        result[k][k], result[k + half][k] = result[k + half][k], result[k][k]

    def double_even():
        c1 = 1
        c2 = n * n
        for i in range(n):
            for j in range(n):
                if i % 4 == j % 4 or (i + j) % 4 == 3:
                    result[i][j] = c2
                else:
                    result[i][j] = c1
                c2 -= 1
                c1 += 1

    if n & 1:
        odd()
    elif n & 3:
        single_even()
    else:
        double_even()
    return result


def levenshtein(x: ite, y: ite) -> int:
    """
    Introduction
    ==========
    Calculate the Levenshtein distance between two sequences.

    The Levenshtein distance is a measure of the difference between two sequences. It is defined as the minimum number
    of single-character edits (i.e., insertions, deletions, or substitutions) required to change one sequence into the
    other. This function supports any type of iterable sequences, such as strings, lists, or tuples.

    Example
    ==========
    >>> levenshtein("kitten", "sitting")
    3
    >>>
    :param x: First sequence to compare.
    :param y: Second sequence to compare.
    :return: The Levenshtein distance between the two sequences.
    """
    if len(x) > len(y):
        x, y = y, x
    previous_row = list(range(len(x) + 1))
    for i2, c2 in enumerate(y):
        current_row = [i2 + 1]
        for i1, c1 in enumerate(x):
            insertions = previous_row[i1 + 1] + 1
            deletions = current_row[i1] + 1
            substitutions = previous_row[i1] + (c1 != c2)
            current_row.append(min(insertions, deletions, substitutions))
        previous_row = current_row
    return previous_row[-1]


def lcsubseq(x: ite, y: ite) -> list:
    """
    Introduction
    ==========
    Calculate the Longest Common Subsequence (LCS) between two sequences.

    The Longest Common Subsequence is a sequence that appears in both given sequences in the same order, but not
    necessarily consecutively. This function uses dynamic programming to build a table that stores the lengths of the
    longest common subsequences of substring pairs, and then backtracks to construct the LCS.

    Example
    ==========
    >>> lcsubseq("ABCBDAB", "BDCAB")
    ['B', 'D', 'A', 'B']
    >>>
    :param x: First sequence to compare.
    :param y: Second sequence to compare.
    :return: The Longest Common Subsequence between the two sequences.
    """
    m, n = len(x), len(y)
    dp = [[0] * (n + 1) for _ in range(m + 1)]
    for i in range(1, m + 1):
        for j in range(1, n + 1):
            if x[i - 1] == y[j - 1]:
                dp[i][j] = dp[i - 1][j - 1] + 1
            else:
                dp[i][j] = max(dp[i - 1][j], dp[i][j - 1])
    i, j = m, n
    seq = []
    while i > 0 and j > 0:
        if x[i - 1] == y[j - 1]:
            seq.append(x[i - 1])
            i -= 1
            j -= 1
        elif dp[i - 1][j] > dp[i][j - 1]:
            i -= 1
        else:
            j -= 1
    seq.reverse()
    return seq


def lcsubstr(x: ite, y: ite) -> list:
    """
    Introduction
    ==========
    Calculate the Longest Common Substring between two sequences.

    The Longest Common Substring is a substring that appears in both given sequences in the same order and is
    consecutive. This function uses dynamic programming to build a table that stores the lengths of the longest common
    substrings of substring pairs, and keeps track of the maximum length and its ending index to extract the substring.

    Example
    ==========
    >>> lcsubstr("ABCBDAB", "BDCAB")
    ['A', 'B']
    >>>
    :param x: First sequence to compare.
    :param y: Second sequence to compare.
    :return: The Longest Common Substring between the two sequences.
    """
    m, n = len(x), len(y)
    dp = [[0] * (n + 1) for _ in range(m + 1)]
    max_len = 0
    end_idx = 0
    for i in range(1, m + 1):
        for j in range(1, n + 1):
            if x[i - 1] == y[j - 1]:
                dp[i][j] = dp[i - 1][j - 1] + 1
                if dp[i][j] > max_len:
                    max_len = dp[i][j]
                    end_idx = i
            else:
                dp[i][j] = 0
    return list(x[end_idx - max_len:end_idx])


def fast_pow(a: Any, n: int, init: Any = 1, mul: Callable = None) -> Any:
    """
    Introduction
    ==========
    This function computes the power of a given base using the fast power algorithm.

    It is a versatile tool that can be used with any data type supporting multiplication, including the ability to
    perform matrix exponentiation by setting an initial matrix and a matrix multiplication function.

    For instance, to compute the matrix power of a 2x2 matrix, you can define the matrix as the base 'a',
    set an identity matrix as the 'init', and provide a function for matrix multiplication as 'mul'.

    Example
    ==========
    >>> fast_pow([[1, 2], [3, 4]], 31, init=[[1, 0], [0, 1]], mul=lambda _a, _b:
    ... [[_a[0][0] * _b[0][0] + _a[0][1] * _b[1][0], _a[0][0] * _b[0][1] + _a[0][1] * _b[1][1]],
    ... [_a[1][0] * _b[0][0] + _a[1][1] * _b[1][0], _a[1][0] * _b[0][1] + _a[1][1] * _b[1][1]]])
    [[10306408007805049875493, 15020838414172076413286], [22531257621258114619929, 32837665629063164495422]]
    >>>
    :param a: The base of the exponentiation, which can be any type of data.
    :param n: The exponent, which must be a non-negative integer.
    :param init: The initial value, which defaults to 1. It is used as the starting point for the computation.
    :param mul: A custom multiplication operation, which defaults to None. If provided, it should be a callable
                that takes two arguments and returns their product.
    :return: The result of the exponentiation, which will be of the same type as the 'init' parameter.
    """
    if mul is None:
        def mul(_a, _b):
            return _a * _b
    res = init
    while n:
        if n & 1:
            res = mul(res, a)
        a = mul(a, a)
        n >>= 1
    return res


if __name__ == "__main__":
    import doctest

    doctest.testmod(verbose=True)



from .types import arr


def lin_reg(x: arr, y: arr) -> list:
    from .maths import mean
    if len(x) != len(y):
        raise ValueError("The array length of the independent and dependent variables must be equal")
    x_mean = mean(x)
    y_mean = mean(y)
    numerator = sum([(x - x_mean) * (y - y_mean) for x, y in zip(x, y)])
    denominator = sum([(x - x_mean) ** 2 for x in x])
    alpha = numerator / denominator
    beta = y_mean - alpha * x_mean
    return [alpha, beta]


def par_reg(x: arr, y: arr) -> list:
    from .maths import mean, var
    if len(x) != len(y):
        raise ValueError("The array length of the independent and dependent variables must be equal")

    def solve_equations(a1, b1, c1, a2, b2, c2):
        denominator = a1 * b2 - a2 * b1
        if denominator == 0:
            return None
        return [(c1 * b2 - c2 * b1) / denominator, (a1 * c2 - a2 * c1) / denominator]

    x2 = [_ ** 2 for _ in x]
    xy = [a + b for a, b in zip(x, y)]
    x2x = [a ** 2 + a for a in x]
    x2y = [a ** 2 + b for a, b in zip(x, y)]
    vx2 = var(x2)
    vxy = var(xy)
    vx2x = var(x2x)
    vx2y = var(x2y)
    v1y = (vxy - var(x) - var(y)) / 2
    v2y = (vx2y - vx2 - var(y)) / 2
    v12 = (vx2x - vx2 - var(x)) / 2
    o = [var(x), v12, v1y, v12, vx2, v2y]
    b, a = solve_equations(o[0], o[1], o[2], o[3], o[4], o[5])
    c = mean(y) - b * mean(x) - a * mean([_ ** 2 for _ in x])
    return [a, b, c]


def poly_reg(x: arr, y: arr, n: int = None) -> list:
    from .matrices import mat
    if len(x) != len(y):
        raise ValueError("The array length of the independent and dependent variables must be equal")
    if n is None:
        n = len(x) - 1
    if not isinstance(n, int) or n <= 0:
        raise ValueError("The degree of a polynomial must be a natural number")
    m = len(x)
    _x = [[_ ** (n - i) for _ in x] for i in range(n)]
    _x.append([1] * m)
    _x = mat(_x)
    return ((_x @ _x.t()).inv() @ _x @ mat([list(y)]).t()).flatten()



from random import choice as __choice, gauss as __gauss, randint as __randint, random as __random, uniform as __uniform
from .types import RandomError, Union, arr, ite, real


def __validate_shape(shape):
    if not (isinstance(shape, (list, tuple)) and all([isinstance(item, int) and item > 0 for item in shape])):
        raise RandomError("The shape must be all positive integers")


def __create_nested_list(dimensions, func):
    return func() if len(dimensions) == 0 else [__create_nested_list(dimensions[1:], func)
                                                for _ in range(dimensions[0])]


def choice(seq: ite, shape: arr = None):
    if not isinstance(seq, (list, tuple, str)):
        raise TypeError("The parameter seq must be iterable")
    if shape is not None:
        __validate_shape(shape)
        return __create_nested_list(shape, lambda: __choice(seq))
    return __choice(seq)


def gauss(mu: real = 0, sigma: real = 1, shape: arr = None) -> Union[float, list]:
    if not (isinstance(mu, (int, float)) and isinstance(sigma, (int, float))):
        raise RandomError("The parameters mu and sigma must both be real numbers")
    if shape is not None:
        __validate_shape(shape)
        return __create_nested_list(shape, lambda: __gauss(mu, sigma))
    return __gauss(mu, sigma)


def randint(a: int, b: int, shape: arr = None) -> Union[int, list]:
    if not (isinstance(a, int) and isinstance(b, int)):
        raise RandomError("The range must be all integers")
    if shape is not None:
        __validate_shape(shape)
        return __create_nested_list(shape, lambda: __randint(a, b))
    return __randint(a, b)


def rand(shape: arr = None) -> Union[float, list]:
    if shape is not None:
        __validate_shape(shape)
        return __create_nested_list(shape, __random)
    return __random()


def uniform(a: real, b: real, shape: arr = None) -> Union[float, list]:
    if not (isinstance(a, (int, float)) and isinstance(b, (int, float))):
        raise RandomError("The range must be all real numbers")
    if shape is not None:
        __validate_shape(shape)
        return __create_nested_list(shape, lambda: __uniform(a, b))
    return __uniform(a, b)



import warnings
import zlib
from .matrices import mat

MODE = {"GRAY": 0, "RGB": 2, "INDEX": 3, "GRAYA": 4, "RGBA": 6}
NUMBER = {0: "GRAY", 2: "RGB", 3: "INDEX", 4: "GRAYA", 6: "RGBA"}
TABLE = [
    0, 1996959894, 3993919788, 2567524794, 124634137, 1886057615, 3915621685, 2657392035,
    249268274, 2044508324, 3772115230, 2547177864, 162941995, 2125561021, 3887607047, 2428444049,
    498536548, 1789927666, 4089016648, 2227061214, 450548861, 1843258603, 4107580753, 2211677639,
    325883990, 1684777152, 4251122042, 2321926636, 335633487, 1661365465, 4195302755, 2366115317,
    997073096, 1281953886, 3579855332, 2724688242, 1006888145, 1258607687, 3524101629, 2768942443,
    901097722, 1119000684, 3686517206, 2898065728, 853044451, 1172266101, 3705015759, 2882616665,
    651767980, 1373503546, 3369554304, 3218104598, 565507253, 1454621731, 3485111705, 3099436303,
    671266974, 1594198024, 3322730930, 2970347812, 795835527, 1483230225, 3244367275, 3060149565,
    1994146192, 31158534, 2563907772, 4023717930, 1907459465, 112637215, 2680153253, 3904427059,
    2013776290, 251722036, 2517215374, 3775830040, 2137656763, 141376813, 2439277719, 3865271297,
    1802195444, 476864866, 2238001368, 4066508878, 1812370925, 453092731, 2181625025, 4111451223,
    1706088902, 314042704, 2344532202, 4240017532, 1658658271, 366619977, 2362670323, 4224994405,
    1303535960, 984961486, 2747007092, 3569037538, 1256170817, 1037604311, 2765210733, 3554079995,
    1131014506, 879679996, 2909243462, 3663771856, 1141124467, 855842277, 2852801631, 3708648649,
    1342533948, 654459306, 3188396048, 3373015174, 1466479909, 544179635, 3110523913, 3462522015,
    1591671054, 702138776, 2966460450, 3352799412, 1504918807, 783551873, 3082640443, 3233442989,
    3988292384, 2596254646, 62317068, 1957810842, 3939845945, 2647816111, 81470997, 1943803523,
    3814918930, 2489596804, 225274430, 2053790376, 3826175755, 2466906013, 167816743, 2097651377,
    4027552580, 2265490386, 503444072, 1762050814, 4150417245, 2154129355, 426522225, 1852507879,
    4275313526, 2312317920, 282753626, 1742555852, 4189708143, 2394877945, 397917763, 1622183637,
    3604390888, 2714866558, 953729732, 1340076626, 3518719985, 2797360999, 1068828381, 1219638859,
    3624741850, 2936675148, 906185462, 1090812512, 3747672003, 2825379669, 829329135, 1181335161,
    3412177804, 3160834842, 628085408, 1382605366, 3423369109, 3138078467, 570562233, 1426400815,
    3317316542, 2998733608, 733239954, 1555261956, 3268935591, 3050360625, 752459403, 1541320221,
    2607071920, 3965973030, 1969922972, 40735498, 2617837225, 3943577151, 1913087877, 83908371,
    2512341634, 3803740692, 2075208622, 213261112, 2463272603, 3855990285, 2094854071, 198958881,
    2262029012, 4057260610, 1759359992, 534414190, 2176718541, 4139329115, 1873836001, 414664567,
    2282248934, 4279200368, 1711684554, 285281116, 2405801727, 4167216745, 1634467795, 376229701,
    2685067896, 3608007406, 1308918612, 956543938, 2808555105, 3495958263, 1231636301, 1047427035,
    2932959818, 3654703836, 1088359270, 936918000, 2847714899, 3736837829, 1202900863, 817233897,
    3183342108, 3401237130, 1404277552, 615818150, 3134207493, 3453421203, 1423857449, 601450431,
    3009837614, 3294710456, 1567103746, 711928724, 3020668471, 3272380065, 1510334235, 755167117
]


def crc(data, length=None, init=4294967295, xor=4294967295):
    if length is None:
        length = len(data)
    for n in range(length):
        c = init ^ data[n]
        init = (TABLE[c & 0xff] ^ (init >> 8)) & 4294967295
    return init ^ xor


class PNG:
    """
    Introduction
    ==========
    This is a PNG class written in pure Python,
    supporting the creation, reading, modification, and saving of PNG images.

    Roadmap
    ==========
    This class is currently in development.
    Future updates will expand its functionality to enhance the capabilities of working with PNG images.

    Usage
    ==========

    Creating a new PNG image:
    ----------
    To create a new PNG image, instantiate the PNG class and use the `new` method
    to define the image dimensions, bit depth, and color mode.

    - from png import PNG

    # Create a new image with a width of 200 pixels, a height of 100 pixels,
    an 8-bit depth, and the default RGB color mode.

    - image = PNG()
    - image.new(200, 100, 8)

    # Optionally, you can specify a background color. For example, to create a new image with a blue background:

    - image.new(200, 100, 8, color=(0, 0, 255))

    Reading an existing PNG image:
    ----------
    To read an existing PNG image from a file, use the `read` method.

    - image = PNG()
    - image.read("example.png")

    Modifying a pixel:
    ----------
    To modify the color of a pixel at a specific coordinate, use the `setp` method.

    # Set the pixel at (10, 10) to red (255, 0, 0).

    - image.setp(10, 10, (255, 0, 0))

    Getting a pixel's color:
    ----------
    To retrieve the color of a pixel at a specific coordinate, use the `getp` method.

    - color = image.getp(10, 10)
    - print(color)  # Output: (255, 0, 0) for the example above

    Saving the image:
    ----------
    To save the image to a file, use the `write` method.

    - image.write("output.png")

    Getting image information:
    ----------
    To obtain information about the image, such as its dimensions and color mode, use the `info` method.

    - info = image.info()
    - print(info)  # Output: {'width': 200, 'height': 100, 'bit_depth': 8, 'color_mode': 'RGB'}
    """

    def __init__(self) -> None:
        self.__bit_depth = None
        self.__color_mode = None
        self.__width = None
        self.__height = None
        self.__pixels = None

    def new(self, width: int, height: int, bit_depth: int, color: tuple = (), color_mode: str = "RGB") -> None:
        color_mode = color_mode.strip().upper()
        supported = ["RGB", "RGBA"]
        maximum = (1 << bit_depth) - 1
        if color_mode not in supported:
            raise ValueError("The current version supports only one color mode from {}".format(
                ", ".join(map(repr, supported))))
        if bit_depth not in [8, 16]:
            raise ValueError("The bit depth of a single channel must be 8 or 16")
        if color:
            color = tuple(color)
        else:
            color = (maximum,) * len(color_mode)
        if len(color) != len(color_mode):
            raise ValueError("The specified color can only be an {} tuple".format(color_mode))
        if not all([isinstance(_, int) and 0 <= _ <= maximum for _ in color]):
            raise ValueError("The color value must be an integer between 0 and {} when the bit depth is {}".format(
                maximum, bit_depth))
        self.__width = width
        self.__height = height
        self.__pixels = mat([[color for _ in range(width)] for _ in range(height)])
        self.__bit_depth = bit_depth
        self.__color_mode = color_mode

    def read(self, filename: str) -> None:
        with open(filename, "rb") as rb:
            part = rb.read(8)
            if part != b"\x89PNG\r\n\x1a\n":
                raise ValueError("Not a PNG file")
            temp = []
            part = rb.read(4)
            length = int.from_bytes(part, "big")
            part = rb.read(4)
            temp.append(part)
            if part != b"IHDR":
                raise ValueError("Missing IHDR chunk")
            if length != 13:
                raise ValueError("The length of IHDR chunk must be 13")
            width, height = rb.read(4), rb.read(4)
            self.__width, self.__height = int.from_bytes(width, "big"), int.from_bytes(height, "big")
            part = rb.read(5)
            temp.append(width)
            temp.append(height)
            temp.append(part)
            self.__bit_depth = part[0]
            if self.__bit_depth not in [8, 16]:
                warnings.warn("Only supports parsing PNG files with a depth of 8 or 16 bits and true color",
                              RuntimeWarning)
            self.__color_mode = NUMBER[part[1]]
            part = rb.read(4)
            if crc(b"".join(temp)) != int.from_bytes(part, "big"):
                warnings.warn("IHDR chunk verification failed", RuntimeWarning)
            while True:
                part = rb.read(4)
                length = int.from_bytes(part, "big")
                part = rb.read(4)
                if part == b"IDAT":
                    pixels = rb.read(length)
                    break
                else:
                    rb.read(length)
                    rb.read(4)
            part = rb.read(4)
            if crc(b"IDAT" + pixels) != int.from_bytes(part, "big"):
                warnings.warn("IDAT chunk verification failed", RuntimeWarning)
            while True:
                part = rb.read(4)
                length = int.from_bytes(part, "big")
                part = rb.read(4)
                if part == b"IEND":
                    part = rb.read(4)
                    if part != b"\xaeB`\x82":
                        warnings.warn("IEND chunk verification failed", RuntimeWarning)
                    break
                else:
                    rb.read(length)
                    rb.read(4)
            de = zlib.decompress(pixels)
            length = len(de)
            line = length // self.__height
            step = len(self.__color_mode)
            if self.__bit_depth == 8:
                self.__pixels = mat([[tuple(de[y + x:y + x + step]) for x in range(1, line, step)]
                                     for y in range(0, length, line)])
            elif self.__bit_depth == 16:
                step *= 2
                self.__pixels = mat([[tuple([int.from_bytes(de[y + x + i:y + x + i + 2], "big")
                                             for i in range(0, step, 2)]) for x in range(1, line, step)]
                                     for y in range(0, length, line)])

    def write(self, filename: str = None) -> bytes:
        if self.__pixels is None:
            raise ValueError("Please use the 'new' method to create an image first")
        ihdr = b"".join([b"IHDR", self.__width.to_bytes(4, "big"), self.__height.to_bytes(4, "big"),
                         bytes([self.__bit_depth]), bytes([MODE[self.__color_mode]]), b"\x00\x00\x00"])

        def func(x):
            return b"".join([_.to_bytes(self.__bit_depth // 8, "big") for _ in x])

        if self.__bit_depth == 8:
            func = bytes
        idat = b"IDAT" + zlib.compress(b"".join([b"\x00" + b"".join(map(func, row)) for row in self.__pixels]))
        png = b"".join([
            b"\x89PNG\r\n\x1a\n", b"\x00\x00\x00\r", ihdr, crc(ihdr).to_bytes(4, "big"),
            b"\x00\x00\x00\x01sRGB\x00\xae\xce\x1c\xe9", b"\x00\x00\x00\x04gAMA\x00\x00\xb1\x8f\x0b\xfca\x05",
            b"\x00\x00\x00\tpHYs\x00\x00\x0e\xc3\x00\x00\x0e\xc3\x01\xc7o\xa8d",
            (len(idat) - 4).to_bytes(4, "big"), idat, crc(idat).to_bytes(4, "big"),
            b"\x00\x00\x00\x00IEND\xaeB`\x82"
        ])
        if filename:
            with open(filename, "wb") as wb:
                wb.write(png)
        else:
            return png

    def getp(self, x: int, y: int) -> tuple:
        return self.__pixels[y, x]

    def setp(self, x: int, y: int, color: tuple) -> None:
        color = tuple(color)
        if len(color) != len(self.__color_mode):
            raise ValueError("The specified color can only be an {} tuple".format(self.__color_mode))
        maximum = (1 << self.__bit_depth) - 1
        if not all([isinstance(_, int) and 0 <= _ <= maximum for _ in color]):
            raise ValueError("The color value must be an integer between 0 and {} when the bit depth is {}".format(
                maximum, self.__bit_depth))
        self.__pixels[y, x] = color

    def info(self) -> dict:
        return {"width": self.__width, "height": self.__height,
                "bit_depth": self.__bit_depth, "color_mode": self.__color_mode}

    def __repr__(self) -> str:
        return self.__class__.__name__ + str(sorted(self.info().items())).replace("',", " =").replace("'", "")



from .types import ContentError

ContentError = ContentError("The input string is invalid")
ORD_A = 65
ORD_a = 97
MORSE_CODE = {
    "A": ".-", "B": "-...", "C": "-.-.", "D": "-..", "E": ".", "F": "..-.", "G": "--.", "H": "....", "I": "..",
    "J": ".---", "K": "-.-", "L": ".-..", "M": "--", "N": "-.", "O": "---", "P": ".--.", "Q": "--.-", "R": ".-.",
    "S": "...", "T": "-", "U": "..-", "V": "...-", "W": ".--", "X": "-..-", "Y": "-.--", "Z": "--..", "0": "-----",
    "1": ".----", "2": "..---", "3": "...--", "4": "....-", "5": ".....", "6": "-....", "7": "--...", "8": "---..",
    "9": "----.", ".": ".-.-.-", ",": "--..--", "?": "..--..", "!": "-.-.--", "'": '.----.', "/": "-..-.", "(": "-.--.",
    ")": "-.--.-", "&": ".-...", ":": "---...", ";": "-.-.-.", "=": "-...-", "+": ".-.-.", "-": "-....-", "_": "..--.-",
    "\"": ".-..-.", "$": "...-..-", "@": ".--.-.", " ": "/"
}
MORSE_CODE_REVERSE = {v: k for k, v in MORSE_CODE.items()}


def base_64(text: str, decrypt: bool = False) -> str:
    base64_chars = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/"
    base64_index = {c: i for i, c in enumerate(base64_chars)}
    base64_padding = "="

    def base64_encode(data):
        encoded = ""
        remainder = len(data) % 3
        padding_len = 3 - remainder if remainder else 0
        for i in range(0, len(data) - remainder, 3):
            chunk = data[i:i + 3]
            int_value = (chunk[0] << 16) + (chunk[1] << 8) + chunk[2]
            for j in range(4):
                bits = (int_value >> (6 * (3 - j))) & 0x3F
                encoded += base64_chars[bits]
        if padding_len:
            chunk = data[-remainder:]
            binary = "".join([format(i, "08b") for i in chunk]) + "0" * (6 - len(chunk) * 8 % 6)
            encoded += "".join([base64_chars[int(binary[j:j + 6], 2)] for j in range(
                0, len(binary), 6)]) + base64_padding * (padding_len % 3)
        return encoded

    def base64_decode(encoded):
        encoded = encoded.replace("\n", "").replace(" ", "")
        decoded = b""
        if not all([c in base64_chars or c == base64_padding for c in encoded]):
            raise ContentError
        position = encoded.find(base64_padding)
        padding_len = len(encoded) - position if position != -1 else 0
        encoded = encoded.rstrip(base64_padding)
        if padding_len > 2 or base64_padding in encoded:
            raise ContentError
        idx = 0
        while idx < len(encoded):
            chunk = encoded[idx:idx + 4]
            int_value = 0
            for j in range(4):
                if j < len(chunk):
                    int_value = (int_value << 6) + base64_index[chunk[j]]
            decoded += (int_value >> 16 & 0xFF).to_bytes(1, "big")
            decoded += (int_value >> 8 & 0xFF).to_bytes(1, "big")
            decoded += (int_value & 0xFF).to_bytes(1, "big")
            idx += 4
        if padding_len:
            try:
                decoded = decoded[:-3] + int(format(int.from_bytes(
                    decoded[-3:], "big") >> 2 * padding_len, "b"), 2).to_bytes(3 - padding_len, "big")
            except OverflowError:
                raise ContentError
        return decoded

    if decrypt:
        decoded_bytes = base64_decode(text).decode("UTF-8")
        return decoded_bytes
    else:
        encoded_bytes = base64_encode(text.encode("UTF-8"))
        return encoded_bytes


def atbash(text: str) -> str:
    return "".join([chr(219 - ord(c)) if "a" <= c <= "z"
                    else chr(155 - ord(c)) if "A" <= c <= "Z" else c for c in text])


def rot13(text: str) -> str:
    return caesar(text, 13)


def caesar(text: str, shift: int, decrypt: bool = False) -> str:
    result = ""
    if decrypt:
        shift = -shift
    for char in text:
        if char.isalpha():
            ascii_offset = ORD_a if char.islower() else ORD_A
            new_char = chr((ord(char) - ascii_offset + shift) % 26 + ascii_offset)
            result += new_char
        else:
            result += char
    return result


def vigenere(text: str, key: str, decrypt: bool = False) -> str:
    result = ""
    key_index = 0
    for char in text:
        if char.isalpha():
            key_char = key[key_index]
            key_index = (key_index + 1) % len(key)
            shift = ord(key_char.lower()) - ORD_a
            if decrypt:
                shift = 26 - shift
            if char.isupper():
                result += chr((ord(char) - ORD_A + shift) % 26 + ORD_A)
            else:
                result += chr((ord(char) - ORD_a + shift) % 26 + ORD_a)
        else:
            result += char
    return result


def substitution(text: str, sub_map: dict, decrypt: bool = False) -> str:
    result = ""
    if decrypt:
        reverse_map = {v: k for k, v in sub_map.items()}
        sub_map = reverse_map
    for char in text:
        if char in sub_map:
            result += sub_map[char]
        else:
            result += char
    return result


def morse(text: str, decrypt: bool = False) -> str:
    code_dict = MORSE_CODE_REVERSE if decrypt else MORSE_CODE
    result = []
    if decrypt:
        text = text.split()
    else:
        text = text.upper()
    for item in text:
        if item in code_dict:
            result.append(code_dict[item])
        else:
            result.append(item)
    char = "" if decrypt else " "
    return char.join(result)


def playfair(text: str, key: str, decrypt: bool = False) -> str:
    def shift(coordinate, direction):
        row, col = coordinate
        if direction == "R":
            col = (col + 1) % 5
        elif direction == "L":
            col = (col - 1) % 5
        elif direction == "D":
            row = (row + 1) % 5
        elif direction == "U":
            row = (row - 1) % 5
        return row, col

    key = key.upper().replace("J", "I")
    dedup = []
    for k in key:
        if k.isalpha() and k not in dedup:
            dedup.append(k)
    key = "".join(dedup)
    if len(text) & 1 != 0:
        text += key[0]
    char_to_coord = {}
    coords = [(i, j) for i in range(5) for j in range(5)]
    letters = ["A", "B", "C", "D", "E", "F", "G", "H", "I", "K", "L", "M", "N", "O", "P", "Q", "R", "S", "T", "U", "V",
               "W", "X", "Y", "Z"]
    for char, coord in zip(key, coords):
        char_to_coord[char] = coord
        letters.remove(char)
    for char, coord in zip(letters, coords[len(key):]):
        char_to_coord[char] = coord
    coord_to_char = {v: k for k, v in char_to_coord.items()}
    result = ""
    for i in range(0, len(text), 2):
        pair = text[i:i + 2]
        is_lowercase1 = pair[0].islower()
        is_lowercase2 = pair[1].islower()
        c1, c2 = pair.upper().replace("J", "I")
        if c1 != c2 and pair.isalpha():
            c1_coord, c2_coord = char_to_coord[c1], char_to_coord[c2]
            if c1_coord[0] == c2_coord[0]:
                c1_coord = shift(c1_coord, "L") if decrypt else shift(c1_coord, "R")
                c2_coord = shift(c2_coord, "L") if decrypt else shift(c2_coord, "R")
            elif c1_coord[1] == c2_coord[1]:
                c1_coord = shift(c1_coord, "U") if decrypt else shift(c1_coord, "D")
                c2_coord = shift(c2_coord, "U") if decrypt else shift(c2_coord, "D")
            else:
                c1_coord, c2_coord = (c1_coord[0], c2_coord[1]), (c2_coord[0], c1_coord[1])
            encrypted_c1, encrypted_c2 = coord_to_char[c1_coord], coord_to_char[c2_coord]
            if is_lowercase1:
                encrypted_c1 = encrypted_c1.lower()
            if is_lowercase2:
                encrypted_c2 = encrypted_c2.lower()
            result += encrypted_c1 + encrypted_c2
        else:
            result += pair
    return result


def hill256(text: bytes, key: list, decrypt: bool = False) -> bytes:
    from .arrays import fill
    from .matrices import mat

    def decrypt_key(k):
        try:
            det = round(k.det())
        except ValueError:
            raise ValueError("The key must be a square matrix")
        res = 0
        while res < mod:
            if det * res % mod == 1:
                break
            else:
                res += 1
                if res == mod:
                    raise ValueError("The key square matrix does not have a multiplicative inverse element")
        return round(k.inv() * det % mod) * res % mod

    if not isinstance(text, bytes):
        raise TypeError("The input content needs to be encoded as a byte type")
    mod = 0x100
    length = len(text)
    key = mat(key)
    if decrypt:
        key = decrypt_key(key)
    cols = key.cols
    result = key @ mat(fill([cols, length // cols + bool(length % cols)], text, False)) % mod
    return b"".join([result[i][j].to_bytes(1, "big") for i in range(result.rows) for j in range(result.cols)])


def ksa(key: bytes) -> list:
    if not isinstance(key, bytes):
        raise TypeError("The key needs to be of byte type")
    length = len(key)
    if not length:
        raise ValueError("The key cannot be empty")
    s = list(range(0x100))
    j = 0
    for i in range(0x100):
        j = (j + s[i] + key[i % length]) % 0x100
        s[i], s[j] = s[j], s[i]
    return s


def prga(s: list):
    i = 0
    j = 0
    while True:
        i = (i + 1) % 0x100
        j = (j + s[i]) % 0x100
        s[i], s[j] = s[j], s[i]
        k = s[(s[i] + s[j]) % 0x100]
        yield k


def rc4(text: bytes, key: bytes) -> bytes:
    if not isinstance(text, bytes):
        raise TypeError("The input content needs to be encoded as a byte type")
    stream = prga(ksa(key))
    result = bytearray()
    for byte in text:
        k = next(stream)
        result.append(byte ^ k)
    return bytes(result)



r"""
 ________   ___    ___  ________   ___    ___  ________    ___  ___   _____ ______
|\   __  \ |\  \  /  /||\   __  \ |\  \  /  /||\   ___  \ |\  \|\  \ |\   _ \  _   \
\ \  \|\  \\ \  \/  / /\ \  \|\  \\ \  \/  / /\ \  \\ \  \\ \  \\\  \\ \  \\\__\ \  \
 \ \   ____\\ \    / /  \ \   ____\\ \    / /  \ \  \\ \  \\ \  \\\  \\ \  \\|__| \  \
  \ \  \___| \/  /  /    \ \  \___| \/  /  /    \ \  \\ \  \\ \  \\\  \\ \  \    \ \  \
   \ \__\  __/  / /       \ \__\  __/  / /       \ \__\\ \__\\ \_______\\ \__\    \ \__\
    \|__| |\___/ /         \|__| |\___/ /         \|__| \|__| \|_______| \|__|     \|__|
          \|___|/                \|___|/

PyPyNum
==========

PyPyNum is a Python library for math & science computations, covering algebra, calculus, stats, with data structures
like matrices, vectors, tensors. It offers numerical tools, programs, and supports computational ops, functions,
processing, simulation, & visualization in data science & ML, crucial for research, engineering, & data processing.

Copyright
==========

- Author: Shen Jiayi
- Email: 2261748025@qq.com
- Copyright: Copyright (c) 2023, Shen Jiayi. All rights reserved.
"""

__author__ = "Shen Jiayi"
__email__ = "2261748025@qq.com"
__copyright__ = "Copyright (c) 2023, Shen Jiayi. All rights reserved."

from . import consts
from . import zh_cn
from .arrays import *
from .chars import int2superscript, superscript2int, int2subscript, subscript2int
from .ciphers import *
from .crandom import *
from .dataproc import *
from .dists import *
from .equations import *
from .fft import *
from .files import *
from .geoms import *
from .graphs import *
from .groups import *
from .images import *
from .interp import *
from .kernels import *
from .logics import *
from .maths import *
from .matrices import *
from .multiprec import *
from .networks import *
from .numbers import *
from .plotting import *
from .polys import *
from .pprinters import *
from .quats import *
from .random import *
from .regs import *
from .seqs import *
from .special import *
from .stattest import *
from .symbols import *
from .tensors import *
from .tools import *
from .trees import *
from .types import config
from .ufuncs import *
from .utils import *
from .vectors import *

__version__ = "1.16.0"
print("PyPyNum", "Version -> " + __version__, "PyPI -> https://pypi.org/project/PyPyNum/",
      "Gitee -> https://www.gitee.com/PythonSJL/PyPyNum", "GitHub -> https://github.com/PythonSJL/PyPyNum", sep=" | ")
for _ in list(globals().keys()):
    if _.endswith("Error"):
        del globals()[_]



"""
A Code Test File
"""

print("""\x1b[92m
from pypynum import (arrays, geoms, logics, matrices, quats, symbols, tensors, vectors,
                     ciphers, consts, equations, maths, plotting, random, regs, tools)
\x1b[m""")
print("...")

from . import (arrays, geoms, logics, matrices, quats, symbols, tensors, vectors,
               ciphers, consts, equations, maths, plotting, random, regs, tools)

# arrays
print("""\x1b[92m
print(arrays.array())
print(arrays.array([1, 2, 3, 4, 5, 6, 7, 8]))
print(arrays.array([[1, 2, 3, 4], [5, 6, 7, 8]]))
print(arrays.array([[[1, 2], [3, 4]], [[5, 6], [7, 8]]]))
\x1b[m""")
print("\"\"\"")
print(arrays.array())
print(arrays.array([1, 2, 3, 4, 5, 6, 7, 8]))
print(arrays.array([[1, 2, 3, 4], [5, 6, 7, 8]]))
print(arrays.array([[[1, 2], [3, 4]], [[5, 6], [7, 8]]]))
print("\"\"\"")

# geoms
print("""\x1b[92m
triangle = geoms.Triangle((0, 0), (2, 2), (3, 0))
print(triangle.perimeter())
print(triangle.area())
print(triangle.centroid())
\x1b[m""")
print("\"\"\"")
triangle = geoms.Triangle((0, 0), (2, 2), (3, 0))
print(triangle.perimeter())
print(triangle.area())
print(triangle.centroid())
print("\"\"\"")

# logics
print("""\x1b[92m
a, b, c = 1, 1, 1
adder0, adder1 = logics.HalfAdder("alpha", a, b), logics.HalfAdder("beta", c, None)
xor0 = logics.XOR("alpha")
ff0, ff1 = logics.DFF("alpha"), logics.DFF("beta")
xor0.set_order0(1)
xor0.set_order1(1)
logics.connector(adder0, adder1)
logics.connector(adder0, xor0)
logics.connector(adder1, xor0)
logics.connector(adder1, ff0)
logics.connector(xor0, ff1)
print("sum: {}, carry: {}".format(ff0.out(), ff1.out()))
\x1b[m""")
print("\"\"\"")
a, b, c = 1, 1, 1
adder0, adder1 = logics.HalfAdder("alpha", a, b), logics.HalfAdder("beta", c, None)
xor0 = logics.XOR("alpha")
ff0, ff1 = logics.DFF("alpha"), logics.DFF("beta")
xor0.set_order0(1)
xor0.set_order1(1)
logics.connector(adder0, adder1)
logics.connector(adder0, xor0)
logics.connector(adder1, xor0)
logics.connector(adder1, ff0)
logics.connector(xor0, ff1)
print("sum: {}, carry: {}".format(ff0.out(), ff1.out()))
print("\"\"\"")

# matrices
print("""\x1b[92m
m0 = matrices.mat([[1, 2], [3, 4]])
m1 = matrices.mat([[5, 6], [7, 8]])
print(m0)
print(m1)
print(m0 + m1)
print(m0 @ m1)
print(m0.inv())
print(m1.rank())
\x1b[m""")
print("\"\"\"")
m0 = matrices.mat([[1, 2], [3, 4]])
m1 = matrices.mat([[5, 6], [7, 8]])
print(m0)
print(m1)
print(m0 + m1)
print(m0 @ m1)
print(m0.inv())
print(m1.rank())
print("\"\"\"")

# quats
print("""\x1b[92m
q0 = quats.quat(1, 2, 3, 4)
q1 = quats.quat(5, 6, 7, 8)
print(q0)
print(q1)
print(q0 + q1)
print(q0 * q1)
print(q0.inverse())
print(q1.conjugate())
\x1b[m""")
print("\"\"\"")
q0 = quats.quat(1, 2, 3, 4)
q1 = quats.quat(5, 6, 7, 8)
print(q0)
print(q1)
print(q0 + q1)
print(q0 * q1)
print(q0.inverse())
print(q1.conjugate())
print("\"\"\"")

# symbols
print("""\x1b[92m
print(symbols.BASIC)
print(symbols.ENGLISH)
print(symbols.GREEK)
print(symbols.parse_expr("-(10+a-(3.14+b0)*(-5))**(-ζn1-2.718/mΣ99)//9"))
\x1b[m""")
print("\"\"\"")
print(symbols.BASIC)
print(symbols.ENGLISH)
print(symbols.GREEK)
print(symbols.parse_expr("-(10+a-(3.14+b0)*(-5))**(-ζn1-2.718/mΣ99)//9"))
print("\"\"\"")

# tensors
print("""\x1b[92m
t0 = tensors.ten([[[1, 2], [3, 4]], [[5, 6], [7, 8]]])
t1 = tensors.ten([[[9, 10], [11, 12]], [[13, 14], [15, 16]]])
print(t0)
print(t1)
print(t0 + t1)
print(t0 @ t1)
\x1b[m""")
print("\"\"\"")
t0 = tensors.ten([[[1, 2], [3, 4]], [[5, 6], [7, 8]]])
t1 = tensors.ten([[[9, 10], [11, 12]], [[13, 14], [15, 16]]])
print(t0)
print(t1)
print(t0 + t1)
print(t0 @ t1)
print("\"\"\"")

# ciphers
print("""\x1b[92m
string = "PyPyNum"
encrypted = ciphers.caesar(string, 10)
print(string)
print(encrypted)
print(ciphers.caesar(encrypted, 10, decrypt=True))
encrypted = ciphers.vigenere(string, "ciphers")
print(string)
print(encrypted)
print(ciphers.vigenere(encrypted, "ciphers", decrypt=True))
encrypted = ciphers.morse(string)
print(string)
print(encrypted)
print(ciphers.morse(encrypted, decrypt=True))
\x1b[m""")
print("\"\"\"")
string = "PyPyNum"
encrypted = ciphers.caesar(string, 10)
print(string)
print(encrypted)
print(ciphers.caesar(encrypted, 10, decrypt=True))
encrypted = ciphers.vigenere(string, "ciphers")
print(string)
print(encrypted)
print(ciphers.vigenere(encrypted, "ciphers", decrypt=True))
encrypted = ciphers.morse(string)
print(string)
print(encrypted)
print(ciphers.morse(encrypted, decrypt=True))
print("\"\"\"")

# vectors
print("""\x1b[92m
v0 = vectors.vec([1, 2, 3, 4])
v1 = vectors.vec([5, 6, 7, 8])
print(v0)
print(v1)
print(v0 + v1)
print(v0 @ v1)
print(v0.normalize())
print(v1.angles())
\x1b[m""")
print("\"\"\"")
v0 = vectors.vec([1, 2, 3, 4])
v1 = vectors.vec([5, 6, 7, 8])
print(v0)
print(v1)
print(v0 * v1)
print(v0 @ v1)
print(v0.normalize())
print(v1.angles())
print("\"\"\"")

# consts
print("""\x1b[92m
print(consts.TB)
print(consts.e)
print(consts.h)
print(consts.phi)
print(consts.pi)
print(consts.tera)
\x1b[m""")
print("\"\"\"")
print(consts.TB)
print(consts.e)
print(consts.h)
print(consts.phi)
print(consts.pi)
print(consts.tera)
print("\"\"\"")

# equations
print("""\x1b[92m
p = [1, -2, -3, 4]
m = [
    [
        [1, 2, 3],
        [6, 10, 12],
        [7, 16, 9]
    ],
    [-1, -2, -3]
]
print(equations.poly_eq(p))
print(equations.lin_eq(*m))
\x1b[m""")
print("\"\"\"")
p = [1, -2, -3, 4]
m = [
    [
        [1, 2, 3],
        [6, 10, 12],
        [7, 16, 9]
    ],
    [-1, -2, -3]
]
print(equations.poly_eq(p))
print(equations.lin_eq(*m))
print("\"\"\"")

# maths
print("""\x1b[92m
print(maths.cot(consts.pi / 3))
print(maths.gamma(1.5))
print(maths.pi(1, 10, lambda x: x ** 2))
print(maths.product([2, 3, 5, 7, 11, 13, 17, 19, 23, 29]))
print(maths.sigma(1, 10, lambda x: x ** 2))
print(maths.var([2, 3, 5, 7, 11, 13, 17, 19, 23, 29]))
\x1b[m""")
print("\"\"\"")
print(maths.cot(consts.pi / 3))
print(maths.gamma(1.5))
print(maths.pi(1, 10, lambda x: x ** 2))
print(maths.product([2, 3, 5, 7, 11, 13, 17, 19, 23, 29]))
print(maths.sigma(1, 10, lambda x: x ** 2))
print(maths.var([2, 3, 5, 7, 11, 13, 17, 19, 23, 29]))
print("\"\"\"")

# plotting
print("""\x1b[92m
plt = plotting.unary(lambda x: x ** 2, top=10, bottom=0, character="+")
print(plt)
print(plotting.binary(lambda x, y: x ** 2 + y ** 2 - 10, right=10, left=0, compare="<=", basic=plotting.change(plt)))
print(plotting.c_unary(lambda x: x ** x, right=2, left=-2, top=2, bottom=-2, complexity=20, character="-"))
\x1b[m""")
print("\"\"\"")
plt = plotting.unary(lambda x: x ** 2, top=10, bottom=0, character="+")
print(plt)
print(plotting.binary(lambda x, y: x ** 2 + y ** 2 - 10, right=10, left=0, compare="<=", basic=plotting.change(plt)))
print(plotting.c_unary(lambda x: x ** x, right=2, left=-2, top=2, bottom=-2, complexity=20, character="-"))
print("\"\"\"")

# random
print("""\x1b[92m
print(random.gauss(0, 1, [2, 3, 4]))
print(random.rand([2, 3, 4]))
print(random.randint(0, 9, [2, 3, 4]))
print(random.uniform(0, 9, [2, 3, 4]))
\x1b[m""")
print("\"\"\"")
print(random.gauss(0, 1, [2, 3, 4]))
print(random.rand([2, 3, 4]))
print(random.randint(0, 9, [2, 3, 4]))
print(random.uniform(0, 9, [2, 3, 4]))
print("\"\"\"")

# regs
print("""\x1b[92m
print(regs.lin_reg(list(range(5)), [2, 4, 6, 7, 8]))
print(regs.par_reg(list(range(5)), [2, 4, 6, 7, 8]))
print(regs.poly_reg(list(range(5)), [2, 4, 6, 7, 8], 4))
\x1b[m""")
print("\"\"\"")
print(regs.lin_reg(list(range(5)), [2, 4, 6, 7, 8]))
print(regs.par_reg(list(range(5)), [2, 4, 6, 7, 8]))
print(regs.poly_reg(list(range(5)), [2, 4, 6, 7, 8], 4))
print("\"\"\"")

# tools
print("""\x1b[92m
print(tools.classify([1, 2.3, 4 + 5j, "string", list, True, 3.14, False, tuple, tools]))
print(tools.dedup(["Python", 6, "NumPy", int, "PyPyNum", 9, "pypynum", "NumPy", 6, True]))
print(tools.frange(0, 3, 0.4))
print(tools.linspace(0, 2.8, 8))
\x1b[m""")
print("\"\"\"")
print(tools.classify([1, 2.3, 4 + 5j, "string", list, True, 3.14, False, tuple, tools]))
print(tools.dedup(["Python", 6, "NumPy", int, "PyPyNum", 9, "pypynum", "NumPy", 6, True]))
print(tools.frange(0, 3, 0.4))
print(tools.linspace(0, 2.8, 8))
print("\"\"\"")

for _ in list(globals().keys()):
    if _[0] != "_":
        del globals()[_]
del _

print("""\x1b[91m
# Tip:
# The test has been successfully passed and ended.
# These tests are only part of the functionality of this package.
# More features need to be explored and tried by yourself!
\x1b[m""")



"""
The Zen of PyPyNum
"""

Zen = ("\x9a¤¢\xa0¢°¼¥°âõòýõäàõÃ°°°°°°°°°°°°°°°°°°°°°°°°°°°°°°°°°°°°°°°°°°°°°°°°°°°°°°°°°°°°°°°°\x9a\x9a"
       "±äù°õóñâòýõ°ã·äõÜ\x9a¾ãõùâÿõøä°üñóùäñýõøäñý°öÿ°þÿùäñêùüñäþõýäâñàýÿó°õøä°ÿä°þùûñ°ãù°þ÷ùãõô°þù°éäùâñüåôÿÝ\x9a"
       "¾ýõâÿõøä°ôõøãùüòñäãõ½üüõç°ñ°öÿ°øäåâä°õøä°øäùç°ãõäñþÿãõâ°äù°¼âñõüó°ãù°þÿùäåüÿã°ñ°öÙ\x9a"
       "¾öÿÿâà°üñóùäñýõøäñý°ñ°öÿ°âÿ÷ùâ°õøä°øäùç°äù°âÿäóñöõâ°¼õæùãåüõ°ãù°þÿùäåüÿã°ñ°öÙ\x9a"
       "¾öÿÿâà°äåÿ½äø÷åÿøä½üüõç°ñ°ãñ°¼ýõüòÿâà°ñ°öÿ°øäàõô°õøä°âõæÿóþå°þñó°õóþõùäñà°¼äõÉ\x9a"
       "¾õæùãùóõô°ôþñ°äöùçã°«öÿÿâà°äóõâùô°õøä°ÿä°þùûñ°ãù°÷þùæüÿã°þù°ããõþäàýÿâÀ\x9a"
       "¾äþõýõüõ°éäùäþõôù°õøä°ãñ°õåáùþå°ãñ°¼þÿùäåüÿã°äþñ÷õüõ°¼õü÷þùã°ñ°õò°ôüåÿøã°õâõøÄ\x9a"
       "¾ýõâÿõøä°ôõþùöõô½üüõç°ñ°öÿ°éäùâñüó°õøä°ûõõã°õç°¼éäùå÷ùòýñ°öÿ°õóñö°õøä°þÙ\x9a"
       "¾óù÷õäñâäã°ôþñ°üñþÿùäþõäþù°ãù°¼ôÿùâõà°ôõóñüà½üüõç°ñ°õûùü°¼õóþõüùÃ\x9a"
       "¾ôõüñõóþÿó°äÿþ°¼ôõãÿàèõ°õò°äãåý°¼öÿÿâà°ñ°þù°þÿùäóùôñâäþÿó°ñ°õûùü°¼ãâÿââÕ\x9a"
       "¾éäùâåà°üñóùäõâÿõøä°ãõôõãâõàåã°þõäöÿ°¼ýõâÿõøä°õüòñóùüààñ°þñ°õûùü°¼éäùüñóùäóñâÀ\x9a"
       "¾õüåâ°üñâõþõ÷°õøä°õäñóãåöòÿ°äÿþ°äãåý°¼öÿÿâà°ñ°þù°ãþÿùäàõóèõ°õûùü°¼ãõãñó°üñùóõàÃ\x9a"
       "¾õüòñãþõàãùôþù°ãù°¼ýõâÿõøä°þõæÿâà½üüõç°ñ°õûùü°¼éäùüùòñôñõÂ\x9a"
       "¾éóþñôþåôõâ°äåÿøäùç°õóþõããõ°ãþùñäõâ°¼üñùýÿþéüÿà°üñýùþùý°ñ°õûùü°¼õôÿó°õãâñàÃ\x9a"
       "¾ãþÿùäóþåö°öÿ°õâåäñþ°âñõþùü°õøä°âÿââùý°õôÿó°âåÿ°þù°ãõùøóâñâõùø°äñüÖ\x9a"
       "¾þÿùäñóùäãùøàÿã°öÿ°õüóñþþùà°õøä°ãù°ýâÿö°þù°éäùóùüàýùÃ\x9a"
       "¾ãýõüòÿâà°äãõûâñô°õøä°ãõäñþùýåüüù°óù÷ÿü°þù°éäùâñüÓ\x9a"
       "¾ãýøäùâÿ÷üñ°éûüåò°õøä°ãôþõóãþñâä°ãóùäñýõøäñý°þù°õóþñ÷õüÕ\x9a¾þÿùäñýùèÿâààñ°ãø÷ùõçäåÿ°þÿùãùóõâÀ\x9a\x9a"
       "¾ãôñõâøä°þÿøäéÀ°õâåà°øäùç°ãýøäùâÿ÷üñ°âåÿ°õæñõç°õç°¼éâñåäóþñã°üñóùäñýõøäñý°ãùøä°þÙ\x9a\x9a"
       "ùéñùÚ°þõøÃ°éò°¼ýåÞéÀéÀ°öÿ°þõÊ°õøÄ°°°°°°°°°°°°°°°°\x9a")
print("".join([chr(ord(_) ^ 144) for _ in Zen[::-1]]))
for _ in list(globals().keys()):
    if _[0] != "_":
        del globals()[_]
del _



from .tools import linspace
from .types import Union, arr, real

thing = Union[list, str]


def color(text: str, rgb: arr) -> str:
    """
    To render colors to text.
    :param text: string
    :param rgb: list | tuple
    :return:
    """
    if not isinstance(rgb, (list, tuple)) or len(rgb) != 3:
        raise ValueError("RGB must be a triplet")
    red, green, blue = rgb
    if not (0 <= red <= 255 and 0 <= green <= 255 and 0 <= blue <= 255):
        raise ValueError("The valid range for RGB values is from 0 to 255")
    number = 16 + round(red / 51.2) * 36 + round(green / 51.2) * 6 + round(blue / 51.2)
    return "\x1b[38;5;{}m{}\x1b[m".format(number, text)


def change(data: thing) -> thing:
    """
    Transform the background between list and string forms.
    :param data: list | string
    :return:
    """
    if isinstance(data, list):
        return "\n".join(["".join(_) for _ in data])
    elif isinstance(data, str):
        return [list(_) for _ in data.split("\n")]
    raise TypeError("The input parameter type can only be a list or string")


def background(right: real = 5, left: real = -5, top: real = 5, bottom: real = -5,
               complexity: real = 5, ratio: real = 3, string: bool = False) -> thing:
    """
    Generate an empty coordinate system.
    :param right: integer | float
    :param left: integer | float
    :param top: integer | float
    :param bottom: integer | float
    :param complexity: integer | float
    :param ratio: integer | float
    :param string: bool.
    :return:
    """
    if abs(ratio) != ratio:
        raise ValueError("The ratio cannot be less than zero")
    if right - left < 1 / complexity or top - bottom < 1 / complexity:
        raise ValueError("The defined width or height cannot be less than the reciprocal of complexity")
    x = linspace(left, right, round((right - left) * complexity + 1))
    plane = [[" "] * 10 + ["|"] + [" "] * len(x) if _ != round((top - bottom) * complexity / 2 / ratio)
             else [" "] * 10 + ["|"] + list(" ".join("_" * (len(x) // 2 + 1))) for _ in range(
        round((top - bottom) * complexity / ratio))] + [[" "] * 10 + ["|"] + ["_"] * len(x)]
    plane[0][:10] = "{:.2e}".format(top).rjust(10)
    plane[-1][:10] = "{:.2e}".format(bottom).rjust(10)
    plane[round((top - bottom) * complexity / 2 / ratio)][:10] = "{:.2e}".format((top + bottom) / 2).rjust(10)
    plane.append([" "] * 11 + list("{:.2e}".format(left).ljust(10)) + list(
        "{:.2e}".format((right + left) / 2).center(len(x) - 20)) + list("{:.2e}".format(right).rjust(10)))
    return plane if not string else change(plane)


def unary(function, right: real = 5, left: real = -5, top: real = 5, bottom: real = -5, complexity: real = 5,
          ratio: real = 3, string: bool = True, basic: list = None, character: str = ".", data: bool = False,
          coloration=False) -> thing:
    """
    Draw a graph of a unary function.
    :param function: function
    :param right: integer | float
    :param left: integer | float
    :param top: integer | float
    :param bottom: integer | float
    :param complexity: integer | float
    :param ratio: integer | float
    :param string: bool.
    :param basic: list
    :param character: string
    :param data: bool.
    :param coloration: bool.
    :return:
    """
    if not isinstance(character, str) or (len(character) != 1 and not coloration):
        raise ValueError("The parameter character must be one character")
    x = linspace(left, right, round((right - left) * complexity + 1))
    y = list(map(function, x))
    plane = [_[:] for _ in basic] if basic else background(right, left, top, bottom, complexity, ratio)
    for i, d in enumerate(y):
        d = round((top - d) * complexity / ratio)
        if 0 <= d <= len(y) - 1:
            plane[d][i + 11] = character
    if string:
        plane = change(plane)
    return [plane, list(zip(x, y))] if data else plane


def binary(function, right: real = 5, left: real = -5, top: real = 5, bottom: real = -5, complexity: real = 5,
           ratio: real = 3, error=0, compare="==", string: bool = True, basic: list = None,
           character: str = ".", data: bool = False, coloration=False) -> thing:
    """
    Draw a graph of binary equations.
    :param function: function
    :param right: integer | float
    :param left: integer | float
    :param top: integer | float
    :param bottom: integer | float
    :param complexity: integer | float
    :param ratio: integer | float
    :param error: integer | float
    :param compare: string
    :param string: bool.
    :param basic: list
    :param character: string
    :param data: bool.
    :param coloration: bool.
    :return:
    """
    if not isinstance(character, str) or (len(character) != 1 and not coloration):
        raise ValueError("The parameter character must be one character")
    x = linspace(left, right, round((right - left) * complexity + 1))
    y = linspace(top, bottom, round((top - bottom) * complexity / ratio + 1))
    plane = [_[:] for _ in basic] if basic else background(right, left, top, bottom, complexity, ratio)
    coordinates = []
    for p1, c1 in enumerate(y):
        for p0, c0 in enumerate(x):
            if compare == "==":
                flag = abs(function(c0, c1)) <= error
            elif compare == "<=":
                flag = function(c0, c1) <= error
            elif compare == ">=":
                flag = function(c0, c1) >= error
            else:
                raise ValueError("The parameter used for comparison can only be '==' or '<=' or '>='")
            if flag:
                if data:
                    coordinates.append((c0, c1))
                plane[p1][p0 + 11] = character
    if string:
        plane = change(plane)
    return [plane, coordinates] if data else plane


def c_unary(function, projection: str = "ri", right: real = 5, left: real = -5, top: real = 5, bottom: real = -5,
            complexity: real = 5, ratio: real = 3, string: bool = True, basic: list = None, character: str = ".",
            data: bool = False, coloration=False) -> thing:
    """
    Draw a graph of a complex function of one variable.
    :param function: function
    :param projection: string
    :param right: integer | float
    :param left: integer | float
    :param top: integer | float
    :param bottom: integer | float
    :param complexity: integer | float
    :param ratio: integer | float
    :param string: bool.
    :param basic: list
    :param character: string
    :param data: bool.
    :param coloration: bool.
    :return:
    """
    if not isinstance(character, str) or (len(character) != 1 and not coloration):
        raise ValueError("The parameter character must be one character")
    x = linspace(left, right, round((right - left) * complexity + 1))
    y = linspace(top, bottom, round((top - bottom) * complexity / ratio + 1))
    plane = [_[:] for _ in basic] if basic else background(right, left, top, bottom, complexity, ratio)
    coordinates = [((c0, c1), function(complex(c0, c1))) for p1, c1 in enumerate(y) for p0, c0 in enumerate(x)]
    for x, y in coordinates:
        if projection == "ri":
            _c0, _c1 = y.real, y.imag
        else:
            raise ValueError("Other modes are currently not supported")
        c0, c1 = round((_c0 - left) * complexity), round((top - _c1) * complexity / ratio)
        if 0 <= c0 <= len(plane[0]) - 12 and 0 <= c1 <= len(plane) - 2:
            plane[c1][c0 + 11] = character
    if string:
        plane = change(plane)
    return [plane, coordinates] if data else plane



from .tools import linspace
from .types import arr


def interp1d(data: arr, length: int) -> list:
    """
    Introduction
    ==========
    One-dimensional data interpolation using piecewise polynomial interpolation.

    Example
    ==========
    >>> interp1d((2, 4, 4, 2), 6)
    [2, 3.320000000000001, 4.160000000000005, 4.160000000000012, 3.3200000000000074, 2]
    >>>
    :param data: List of data points to be interpolated. Must be at least two points.
    :param length: The number of points in the resampled data set.
    :return: A list of interpolated values at the new points.
    """
    from .regs import lin_reg, par_reg
    expr = [lambda x: sum([k * x ** (1 - n) for n, k in enumerate(lin_reg([0, 1], [data[0], data[1]]))])]
    for i in range(len(data) - 2):
        temp = par_reg(list(range(i, i + 3)), data[i:i + 3])
        expr.append(lambda x, coefficients=temp: sum([k * x ** (2 - n) for n, k in enumerate(coefficients)]))
    expr.append(lambda x: sum([k * x ** (1 - n) for n, k in
                               enumerate(lin_reg([len(data) - 2, len(data) - 1], [data[-2], data[-1]]))]))
    result = linspace(0, len(data) - 1, length)
    for item in range(length):
        if int(result[item]) != result[item]:
            result[item] = (expr[int(result[item])](result[item]) + expr[int(result[item] + 1)](result[item])) / 2
        else:
            result[item] = data[int(result[item])]
    return result


def bicubic(x):
    """
    Calculate the bicubic interpolation basis function value.
    :param x: The x value for which the basis function is evaluated.
    :return: The value of the bicubic basis function at x.
    """
    absx = abs(x)
    if absx <= 1:
        return 1.5 * absx ** 3 - 2.5 * absx ** 2 + 1
    elif absx < 2:
        return -0.5 * absx ** 3 + 2.5 * absx ** 2 - 4 * absx + 2
    else:
        return 0


def contribute(src, x, y, channels=None):
    """
    Calculate the contribution of the source array at a specific point using bicubic interpolation.
    :param src: The source 2D array from which to interpolate.
    :param x: The x-coordinate of the point to interpolate.
    :param y: The y-coordinate of the point to interpolate.
    :param channels: The number of channels if src is a multichannel array.
    :return: The interpolated value at the point (x, y).
    """
    src_height = len(src)
    src_width = len(src[0])
    value = 0.0 if channels is None else [0.0] * channels
    x_int = int(x)
    y_int = int(y)
    dx = x - x_int
    dy = y - y_int
    for j in range(-1, 3):
        ny = y_int + j
        if 0 <= ny < src_height:
            wy = bicubic(j - dy)
            for i in range(-1, 3):
                nx = x_int + i
                if 0 <= nx < src_width:
                    wx = bicubic(i - dx)
                    if channels is None:
                        value += wx * wy * src[ny][nx]
                    else:
                        for c in range(channels):
                            value[c] += wx * wy * src[ny][nx][c]
    return value


def interp2d(src, new_height, new_width, channels=None, round_res=False, min_val=None, max_val=None):
    """
    Introduction
    ==========
    Two-dimensional data interpolation using bicubic interpolation.

    Example
    ==========
    >>> interp2d([[1, 2], [3, 4]], 3, 3)
    [[1.0, 1.6875, 2.0], [2.25, 3.1640625, 3.375], [3.0, 3.9375, 4.0]]
    >>>
    :param src: The source 2D array to be interpolated.
    :param new_height: The desired height of the interpolated array.
    :param new_width: The desired width of the interpolated array.
    :param channels: The number of channels if src is a multichannel array.
    :param round_res: Whether to round the result to the nearest integer.
    :param min_val: The minimum value to clip the interpolated results.
    :param max_val: The maximum value to clip the interpolated results.
    :return: A 2D array of the interpolated values with the new dimensions.
    """
    src_height = len(src)
    src_width = len(src[0])
    dst = [[None] * new_width for _ in range(new_height)]
    for dst_y in range(new_height):
        for dst_x in range(new_width):
            src_x = (src_width - 1) * dst_x / (new_width - 1)
            src_y = (src_height - 1) * dst_y / (new_height - 1)
            value = contribute(src, src_x, src_y, channels)
            if min_val is not None:
                value = max(min_val, value) if channels is None else tuple(max(min_val, v) for v in value)
            if max_val is not None:
                value = min(max_val, value) if channels is None else tuple(min(max_val, v) for v in value)
            if round_res:
                value = round(value) if channels is None else tuple(map(round, value))
            dst[dst_y][dst_x] = value
    return dst



from .types import arr, real


class Point:
    def __init__(self, p: arr):
        if isinstance(p, (list, tuple)):
            if len(p) == 2:
                self.p = p
            else:
                raise ValueError("The coordinate value length can only be two")
        else:
            raise TypeError("The type of coordinate value can only be an array")

    def __repr__(self):
        return "Point({})".format(self.p)


class Line:
    def __init__(self, a: arr, b: arr):
        if isinstance(a, (list, tuple)) and isinstance(b, (list, tuple)):
            if len(a) == len(b) == 2:
                self.a, self.b = a, b
            else:
                raise ValueError("The coordinate value length can only be two")
        else:
            raise TypeError("The type of coordinate value can only be an array")

    def __repr__(self) -> str:
        return "Line({}, {})".format(self.a, self.b)

    def length(self) -> float:
        return ((self.a[0] - self.b[0]) ** 2 + (self.a[1] - self.b[1]) ** 2) ** 0.5

    def expr(self) -> list:
        if self.b[0] - self.a[0] == 0:
            k = None
        else:
            k = (self.b[1] - self.a[1]) / (self.b[0] - self.a[0])
        if k is None:
            b = None
        else:
            b = self.a[1] - k * self.a[0]
        return [k, b]

    def vertical(self, p: arr) -> tuple:
        if isinstance(p, (list, tuple)):
            if len(p) == 2:
                k, b = self.expr()
                if k == 0:
                    x = p[0]
                    y = b
                    return x, y
                if k is None:
                    x = self.a[0]
                    y = p[1]
                    return x, y
                kp = -1 / k
                bp = p[1] - kp * p[0]
                x = (bp - b) / (k - kp)
                y = kp * x + bp
                return x, y
            raise ValueError("The coordinate value length can only be two")
        raise TypeError("The type of coordinate value can only be an array")


class Triangle:
    def __init__(self, a: arr, b: arr, c: arr):
        if isinstance(a, (list, tuple)) and isinstance(b, (list, tuple)) and isinstance(c, (list, tuple)):
            if len(a) == len(b) == len(c) == 2:
                self.a, self.b, self.c = a, b, c
            else:
                raise ValueError("The coordinate value length can only be two")
        else:
            raise TypeError("The type of coordinate value can only be an array")

    def __repr__(self) -> str:
        return "Triangle({}, {}, {})".format(self.a, self.b, self.c)

    def length(self) -> tuple:
        return distance(self.a, self.b), distance(self.b, self.c), distance(self.c, self.a)

    def perimeter(self) -> float:
        return sum(self.length())

    def area(self) -> float:
        return abs((self.b[0] - self.a[0]) * (self.c[1] - self.a[1])
                   - (self.b[1] - self.a[1]) * (self.c[0] - self.a[0])) / 2

    def incenter(self) -> tuple:
        a, b, c = self.length()
        x = (a * self.a[0] + b * self.b[0] + c * self.c[0]) / (a + b + c)
        y = (a * self.a[1] + b * self.b[1] + c * self.c[1]) / (a + b + c)
        return x, y

    def circumcenter(self) -> tuple:
        x = (self.a[0] * self.b[0] + self.b[0] * self.c[0] + self.c[0] * self.a[0]) / (
                self.a[0] + self.b[0] + self.c[0])
        y = (self.a[1] * self.b[1] + self.b[1] * self.c[1] + self.c[1] * self.a[1]) / (
                self.a[1] + self.b[1] + self.c[1])
        return x, y

    def centroid(self) -> tuple:
        x = (self.a[0] + self.b[0] + self.c[0]) / 3
        y = (self.a[1] + self.b[1] + self.c[1]) / 3
        return x, y

    def orthocenter(self) -> tuple:
        if self.area() == 0:
            raise ValueError("A triangle with zero area has no orthogonal center")
        a, b, c = self.a, self.b, self.c
        divisor = a[0] * b[1] - a[0] * c[1] - a[1] * b[0] + a[1] * c[0] + b[0] * c[1] - b[1] * c[0]
        x = (-a[0] * a[1] * b[0] + a[0] * a[1] * c[0] + a[0] * b[0] * b[1] - a[0] * c[0] * c[1] - a[1] ** 2 * b[1] + a[
            1] ** 2 * c[1] + a[1] * b[1] ** 2 - a[1] * c[1] ** 2 - b[0] * b[1] * c[0] + b[0] * c[0] * c[1] - b[1] ** 2 *
             c[1] + b[1] * c[1] ** 2) / divisor
        y = (a[0] ** 2 * b[0] - a[0] ** 2 * c[0] + a[0] * a[1] * b[1] - a[0] * a[1] * c[1] - a[0] * b[0] ** 2 + a[0] *
             c[0] ** 2 - a[1] * b[0] * b[1] + a[1] * c[0] * c[1] + b[0] ** 2 * c[0] + b[0] * b[1] * c[1] - b[0] * c[
                 0] ** 2 - b[1] * c[0] * c[1]) / divisor
        return x, y

    def is_isosceles(self, error: real = 0) -> bool:
        _length = self.length()
        return True if abs(_length[0] - _length[1]) <= error or abs(_length[1] - _length[2]) <= error or abs(
            _length[2] - _length[0]) <= error else False

    def is_equilateral(self, error: real = 0) -> bool:
        _length = self.length()
        return True if abs(_length[0] - _length[1]) <= error and abs(_length[1] - _length[2]) <= error and abs(
            _length[2] - _length[0]) <= error else False

    def is_right(self, error: real = 0) -> bool:
        _length = sorted(self.length())
        return True if abs(
            _length[0] * _length[0] + _length[1] * _length[1] - _length[2] * _length[2]) <= error else False


class Quadrilateral:
    def __init__(self, a: arr, b: arr, c: arr, d: arr):
        if isinstance(a, (list, tuple)) and isinstance(b, (list, tuple)) and isinstance(
                c, (list, tuple)) and isinstance(d, (list, tuple)):
            if len(a) == len(b) == len(c) == len(d) == 2:
                self.a, self.b, self.c, self.d = a, b, c, d
            else:
                raise ValueError("The coordinate value length can only be two")
        else:
            raise TypeError("The type of coordinate value can only be an array")

    def __repr__(self) -> str:
        return "Quadrilateral({}, {}, {}, {})".format(self.a, self.b, self.c, self.d)

    def length(self) -> tuple:
        return distance(self.a, self.b), distance(self.b, self.c), distance(self.c, self.d), distance(self.d, self.a)

    def perimeter(self) -> float:
        return sum(self.length())

    def area(self) -> float:
        return abs((self.b[0] - self.a[0]) * (self.c[1] - self.a[1]) - (self.b[1] - self.a[1]) * (
                self.c[0] - self.a[0]) - (self.d[0] - self.a[0]) * (self.c[1] - self.a[1]) + (
                           self.d[1] - self.a[1]) * (self.c[0] - self.a[0])) / 2

    def centroid(self) -> tuple:
        _ = Triangle(self.a, self.b, self.c)
        x1, y1 = _.centroid()
        s1 = _.area()
        _ = Triangle(self.a, self.d, self.c)
        x2, y2 = _.centroid()
        s2 = _.area()
        x = (x1 * s1 + x2 * s2) / (s1 + s2)
        y = (y1 * s1 + y2 * s2) / (s1 + s2)
        return x, y

    def __slopes(self):
        return [Line(self.a, self.b).expr()[0], Line(self.c, self.d).expr()[0], Line(self.b, self.c).expr()[0],
                Line(self.d, self.a).expr()[0]]

    def is_trapezoidal(self, error: real = 0) -> bool:
        slopes = self.__slopes()
        slopes = [_ if _ is not None else 0 for _ in slopes]
        return True if abs(slopes[0] - slopes[1]) <= error or abs(slopes[2] - slopes[3]) <= error else False

    def is_parallelogram(self, error: real = 0) -> bool:
        slopes = self.__slopes()
        slopes = [_ if _ is not None else 0 for _ in slopes]
        return True if abs(slopes[0] - slopes[1]) <= error and abs(slopes[2] - slopes[3]) <= error else False

    def is_diamond(self, error: real = 0) -> bool:
        _length = self.length()
        return True if abs(_length[0] - _length[1]) <= error and abs(_length[1] - _length[2]) <= error and abs(
            _length[2] - _length[3]) <= error and abs(_length[3] - _length[0]) <= error else False

    def is_rectangular(self, error: real = 0) -> bool:
        slopes = [Line(self.a, self.b).expr()[0], Line(self.b, self.c).expr()[0]]
        if None in slopes:
            slopes.remove(None)
            if slopes:
                if abs(slopes[0]) <= error:
                    return True
                return False
            return False
        return True if abs(1 / max(slopes) + min(slopes)) <= error and self.is_parallelogram(error) else False

    def is_square(self, error: real = 0) -> bool:
        return True if self.is_diamond(error) and self.is_rectangular(error) else False


class Polygon:
    def __init__(self, *p: arr):
        if len(p) < 3:
            raise ValueError("The number of vertices in a polygon cannot be less than three")
        for item in p:
            if isinstance(item, (list, tuple)):
                if len(item) != 2:
                    raise ValueError("The coordinate value length can only be two")
            else:
                raise TypeError("The type of coordinate value can only be an array")
        self.points = p

    def __repr__(self) -> str:
        return "Polygon(" + ", ".join(str(_) for _ in self.points) + ")"

    def length(self) -> tuple:
        return tuple(distance(*item) for item in zip(self.points, self.points[1:] + (self.points[0],)))

    def perimeter(self) -> float:
        return sum(self.length())

    def area(self) -> float:
        _area = 0
        q = self.points[-1]
        for p in self.points:
            _area += p[0] * q[1] - p[1] * q[0]
            q = p
        return abs(_area) / 2

    def centroid(self) -> tuple:
        raise NotImplementedError


class Circle:
    def __init__(self, center: arr, radius: real):
        if isinstance(center, (list, tuple)):
            if len(center) == 2:
                self.center = center
            else:
                raise ValueError("The coordinate value length can only be two")
        else:
            raise TypeError("The type of coordinate value can only be an array")
        if isinstance(radius, (int, float)):
            self.radius = radius
        else:
            raise TypeError("The type of length value can only be real numbers")

    def __repr__(self) -> str:
        return "Circle({}, {})".format(self.center, self.radius)

    def perimeter(self) -> float:
        return 6.283185307179586 * self.radius

    def area(self) -> float:
        return 3.141592653589793 * self.radius * self.radius

    def expr(self) -> list:
        return [self.center[0], self.center[1], self.radius]

    def chord(self, radian: real) -> float:
        from math import sin
        return 2 * self.radius * sin(radian / 2)


geom = Point, Line, Triangle, Quadrilateral, Polygon, Circle


def distance(g1, g2, error: real = 0) -> float:
    if isinstance(g1, (list, tuple)) and isinstance(g2, (list, tuple)):
        return ((g1[0] - g2[0]) ** 2 + (g1[1] - g2[1]) ** 2) ** 0.5
    if isinstance(g1, (list, tuple)):
        g1 = Point(g1)
    if isinstance(g2, (list, tuple)):
        g2 = Point(g2)
    if isinstance(g1, geom) and isinstance(g2, geom):
        if geom.index(type(g2)) < geom.index(type(g1)):
            return distance(g2, g1)
        elif isinstance(g1, Point) and isinstance(g2, Point):
            p1, p2 = g1.p, g2.p
            return ((p1[0] - p2[0]) ** 2 + (p1[1] - p2[1]) ** 2) ** 0.5
        elif isinstance(g1, Point) and isinstance(g2, Line):
            return Line(g1.p, g2.vertical(g1.p)).length()
        elif isinstance(g1, Line) and isinstance(g2, Line):
            e1, e2 = g1.expr(), g2.expr()
            if e1 == e2 == [None, None]:
                return abs(g1.a[0] - g2.a[0])
            elif abs(e1[0] - e2[0]) <= error:
                return abs(e1[1] - e2[1]) / (((e1[0] + e2[0]) / 2) ** 2 + 1) ** 0.5
            return 0
        elif isinstance(g1, Point) and isinstance(g2, Triangle):
            to_vertexes = [distance(g1.p, g2.a), distance(g1.p, g2.b), distance(g1.p, g2.c)]
            points = [Line(g2.a, g2.b).vertical(g1.p), Line(g2.b, g2.c).vertical(g1.p), Line(g2.c, g2.a).vertical(g1.p)]
            to_sides = [distance(g1.p, _) for _ in points]
            minimum = min(min(to_vertexes), min(to_sides))
            if min(to_sides) < min(to_vertexes):
                def determine():
                    index = to_sides.index(min(to_sides))
                    if (index == 0 and min(g2.a[0], g2.b[0]) < points[index][0] < max(g2.a[0], g2.b[0])
                        and min(g2.a[1], g2.b[1]) < points[index][1] < max(g2.a[1], g2.b[1])) or (
                            index == 1 and min(g2.b[0], g2.c[0]) < points[index][0] < max(g2.b[0], g2.c[0])
                            and min(g2.b[1], g2.c[1]) < points[index][1] < max(g2.b[1], g2.c[1])) or (
                            index == 2 and min(g2.c[0], g2.a[0]) < points[index][0] < max(g2.c[0], g2.a[0])
                            and min(g2.c[1], g2.a[1]) < points[index][1] < max(g2.c[1], g2.a[1])):
                        return True
                    return False

                if determine():
                    return minimum
                else:
                    to_sides[to_sides.index(min(to_sides))] = float("inf")
                    minimum = min(min(to_vertexes), min(to_sides))
                    if determine():
                        return minimum
                    to_sides[to_sides.index(min(to_sides))] = float("inf")
                    minimum = min(min(to_vertexes), min(to_sides))
            return minimum
        elif isinstance(g1, Line) and isinstance(g2, Triangle):
            pass
        elif isinstance(g1, Triangle) and isinstance(g2, Triangle):
            ...
        raise NotImplementedError
    raise TypeError("The input parameters can only be geometric shapes")
